import re
from textwrap import dedent
from unittest import TestCase

from pcs.cli import nvset
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.types import (
    CibRuleExpressionType,
    CibRuleInEffectStatus,
)


def fixture_dto(in_effect):
    return CibNvsetDto(
        f"id-{in_effect}",
        {"score": "150"},
        CibRuleExpressionDto(
            f"id-{in_effect}-rule",
            CibRuleExpressionType.RULE,
            in_effect,
            {"boolean-op": "or"},
            None,
            None,
            [
                CibRuleExpressionDto(
                    f"id-{in_effect}-rule-op",
                    CibRuleExpressionType.OP_EXPRESSION,
                    CibRuleInEffectStatus.UNKNOWN,
                    {"name": "monitor"},
                    None,
                    None,
                    [],
                    "op monitor",
                ),
            ],
            "op monitor",
        ),
        [CibNvpairDto(f"id-{in_effect}-pair1", "name1", "value1")],
    )


def fixture_dto_list():
    return [fixture_dto(in_effect.value) for in_effect in CibRuleInEffectStatus]


class FilterOutExpiredNvset(TestCase):
    def test_filter(self):
        self.maxDiff = None
        list_with_expired_nvsets = [
            CibNvsetDto(id="nvset-no-rule", options={}, rule=None, nvpairs=[]),
            CibNvsetDto(
                id="nvset-2",
                options={},
                rule=CibRuleExpressionDto(
                    id="rule-another-expired",
                    type=CibRuleExpressionType.DATE_EXPRESSION,
                    in_effect=CibRuleInEffectStatus.EXPIRED,
                    options={},
                    date_spec=None,
                    duration=None,
                    expressions=[],
                    as_string="",
                ),
                nvpairs=[],
            ),
        ] + fixture_dto_list()
        expected_list = [
            item
            for idx, item in enumerate(list_with_expired_nvsets[:])
            if idx not in [1, 4]
        ]
        self.assertEqual(
            nvset.filter_out_expired_nvset(list_with_expired_nvsets),
            expected_list,
        )

    def test_empty_list(self):
        self.assertEqual(nvset.filter_out_expired_nvset([]), [])


class FilterNvpairsByNames(TestCase):
    test_nvsets = [
        CibNvsetDto(
            id="1",
            options={},
            rule=None,
            nvpairs=[
                CibNvpairDto(id="1a", name="a", value="1"),
                CibNvpairDto(id="1b", name="b", value="2"),
                CibNvpairDto(id="1c", name="c", value="3"),
            ],
        ),
        CibNvsetDto(
            id="2",
            options={},
            rule=None,
            nvpairs=[
                CibNvpairDto(id="2a", name="a", value="1"),
                CibNvpairDto(id="2b", name="b", value="2"),
            ],
        ),
    ]

    def test_filter_no_match(self):
        self.assertEqual(
            nvset.filter_nvpairs_by_names(self.test_nvsets, ["x"]),
            [
                CibNvsetDto(id="1", options={}, rule=None, nvpairs=[]),
                CibNvsetDto(id="2", options={}, rule=None, nvpairs=[]),
            ],
        )

    def test_filter_match_in_one_set(self):
        self.assertEqual(
            nvset.filter_nvpairs_by_names(self.test_nvsets, ["c"]),
            [
                CibNvsetDto(
                    id="1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="1c", name="c", value="3"),
                    ],
                ),
                CibNvsetDto(id="2", options={}, rule=None, nvpairs=[]),
            ],
        )

    def test_filter_match_in_more_set(self):
        self.assertEqual(
            nvset.filter_nvpairs_by_names(self.test_nvsets, ["a"]),
            [
                CibNvsetDto(
                    id="1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="1a", name="a", value="1"),
                    ],
                ),
                CibNvsetDto(
                    id="2",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="2a", name="a", value="1"),
                    ],
                ),
            ],
        )

    def test_filter_multiple_filter(self):
        self.assertEqual(
            nvset.filter_nvpairs_by_names(self.test_nvsets, ["a", "c", "x"]),
            [
                CibNvsetDto(
                    id="1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="1a", name="a", value="1"),
                        CibNvpairDto(id="1c", name="c", value="3"),
                    ],
                ),
                CibNvsetDto(
                    id="2",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="2a", name="a", value="1"),
                    ],
                ),
            ],
        )


class NvsetDtoToLines(TestCase):
    def setUp(self):
        self.label = "Meta Attributes"

    def _export(self, dto, with_ids):
        return (
            "\n".join(
                nvset.nvset_dto_to_lines(
                    dto, nvset_label=self.label, with_ids=with_ids
                )
            )
            + "\n"
        )

    def assert_lines(self, dto, lines):
        self.assertEqual(
            self._export(dto, True),
            lines,
        )
        self.assertEqual(
            self._export(dto, False),
            re.sub(r" +\(id:.*\)", "", lines),
        )

    def test_minimal(self):
        dto = CibNvsetDto("my-id", {}, None, [])
        output = dedent(
            f"""\
              {self.label}: my-id
            """
        )
        self.assert_lines(dto, output)

    def test_full(self):
        dto = CibNvsetDto(
            "my-id",
            {"score": "150"},
            CibRuleExpressionDto(
                "my-id-rule",
                CibRuleExpressionType.RULE,
                CibRuleInEffectStatus.UNKNOWN,
                {"boolean-op": "or"},
                None,
                None,
                [
                    CibRuleExpressionDto(
                        "my-id-rule-op",
                        CibRuleExpressionType.OP_EXPRESSION,
                        CibRuleInEffectStatus.UNKNOWN,
                        {"name": "monitor"},
                        None,
                        None,
                        [],
                        "op monitor",
                    ),
                ],
                "op monitor",
            ),
            [
                CibNvpairDto("my-id-pair1", "name1", "value1"),
                CibNvpairDto("my-id-pair2", "name 2", "value 2"),
                CibNvpairDto("my-id-pair3", "name=3", "value=3"),
            ],
        )
        output = dedent(
            f"""\
            {self.label}: my-id score=150
              "name 2"="value 2" (id: my-id-pair2)
              name1=value1 (id: my-id-pair1)
              "name=3"="value=3" (id: my-id-pair3)
              Rule: boolean-op=or (id: my-id-rule)
                Expression: op monitor (id: my-id-rule-op)
            """
        )
        self.assert_lines(dto, output)


class NvsetDtoListToLines(TestCase):
    def setUp(self):
        self.label = "Meta Attributes"

    def _export(self, dto, with_ids):
        return (
            "\n".join(
                nvset.nvset_dto_list_to_lines(
                    dto,
                    nvset_label=self.label,
                    with_ids=with_ids,
                )
            )
            + "\n"
        )

    def assert_lines(self, dto, lines):
        self.assertEqual(
            self._export(dto, True),
            lines,
        )
        self.assertEqual(
            self._export(dto, False),
            re.sub(r" +\(id:.*\)", "", lines),
        )

    def test_lines(self):
        self.maxDiff = None
        output = dedent(
            f"""\
            {self.label} (not yet in effect): id-NOT_YET_IN_EFFECT score=150
              name1=value1 (id: id-NOT_YET_IN_EFFECT-pair1)
              Rule (not yet in effect): boolean-op=or (id: id-NOT_YET_IN_EFFECT-rule)
                Expression: op monitor (id: id-NOT_YET_IN_EFFECT-rule-op)
            {self.label}: id-IN_EFFECT score=150
              name1=value1 (id: id-IN_EFFECT-pair1)
              Rule: boolean-op=or (id: id-IN_EFFECT-rule)
                Expression: op monitor (id: id-IN_EFFECT-rule-op)
            {self.label} (expired): id-EXPIRED score=150
              name1=value1 (id: id-EXPIRED-pair1)
              Rule (expired): boolean-op=or (id: id-EXPIRED-rule)
                Expression: op monitor (id: id-EXPIRED-rule-op)
            {self.label}: id-UNKNOWN score=150
              name1=value1 (id: id-UNKNOWN-pair1)
              Rule: boolean-op=or (id: id-UNKNOWN-rule)
                Expression: op monitor (id: id-UNKNOWN-rule-op)
        """
        )
        self.assert_lines(fixture_dto_list(), output)



from dataclasses import dataclass

from pcs.common.host import Destination, PcsKnownHost


@dataclass(frozen=True)
class HostAuthData:
    username: str
    password: str
    dest_list: list[Destination]


@dataclass(frozen=True)
class HostWithTokenAuthData:
    token: str
    dest_list: list[Destination]

    def to_known_host(self, node_name: str) -> PcsKnownHost:
        return PcsKnownHost(node_name, self.token, self.dest_list)



from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import (
    format_cmd_list,
    lines_to_str,
    smart_wrap_text,
)
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    Argv,
    InputModifiers,
    KeyValueParser,
    wait_to_timeout,
)
from pcs.cli.reports.output import deprecation_warning, warn
from pcs.cli.resource import command as resource_cmd
from pcs.cli.resource.common import get_resource_status_msg
from pcs.cli.stonith.common import check_is_stonith
from pcs.cli.stonith.levels.output import (
    stonith_level_config_to_cmd,
    stonith_level_config_to_text,
)
from pcs.common import reports
from pcs.common.str_tools import indent


def config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --output-format - supported formats: text, cmd, json
      * -f CIB file
    """
    output_format = modifiers.get_output_format()
    output = resource_cmd.config_common(lib, argv, modifiers, stonith=True)

    if output_format == OUTPUT_FORMAT_VALUE_JSON:
        # JSON output format does not include fencing levels because it would
        # change the current JSON structure and break existing user tooling
        warn(
            "Fencing levels are not included because this command could only "
            "export stonith configuration previously. This cannot be changed "
            "to avoid breaking existing tooling. To export fencing levels, run "
            "'pcs stonith level config --output-format=json'"
        )
        print(output)
        return

    fencing_topology_dto = lib.fencing_topology.get_config_dto()
    if output_format == OUTPUT_FORMAT_VALUE_CMD:
        # we can look at the output of config_common as one command
        output = format_cmd_list(
            [output, *stonith_level_config_to_cmd(fencing_topology_dto)]
        )
    else:
        text_output = stonith_level_config_to_text(fencing_topology_dto)
        if text_output:
            output += "\n\nFencing Levels:\n" + lines_to_str(
                smart_wrap_text(indent(text_output))
            )

    if output:
        print(output)


def meta(lib: Any, argv: list[str], modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - override editing connection attributes for guest nodes
      * --wait - wait for cluster to reach steady state
      * -f
    """
    modifiers.ensure_only_supported("-f", "--force", "--wait")
    modifiers.ensure_not_mutually_exclusive("-f", "--wait")
    wait_timeout = wait_to_timeout(modifiers.get("--wait"))
    force_flags = []
    if modifiers.get("--force"):
        force_flags.append(reports.codes.FORCE)
    if not argv:
        raise CmdLineInputError()
    resource_id = argv.pop(0)
    check_is_stonith(lib, [resource_id], "pcs resource meta")
    meta_attrs_dict = KeyValueParser(argv).get_unique()

    lib.resource.update_meta(resource_id, meta_attrs_dict, force_flags)

    if wait_timeout >= 0:
        deprecation_warning(reports.messages.ResourceWaitDeprecated().message)
        lib.cluster.wait_for_pcmk_idle(wait_timeout)
        print(get_resource_status_msg(lib, resource_id))



from dataclasses import dataclass
from typing import Sequence

from pcs.common.interface.dto import DataTransferObject
from pcs.common.types import StringSequence


@dataclass(frozen=True)
class CibTagDto(DataTransferObject):
    id: str
    idref_list: StringSequence


@dataclass(frozen=True)
class CibTagListDto(DataTransferObject):
    tags: Sequence[CibTagDto]



# pylint: disable=too-many-lines
from typing import Optional
from unittest import (
    TestCase,
    mock,
)

from lxml import etree
from lxml.etree import _Element

from pcs.common import reports
from pcs.lib.cib import const
from pcs.lib.cib import remove_elements as lib

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)
from pcs_test.tools.fixture_cib import modify_cib
from pcs_test.tools.fixture_crm_mon import complete_state
from pcs_test.tools.misc import read_test_resource
from pcs_test.tools.xml import etree_to_str

EXPECTED_TYPES_FOR_REMOVE = ["constraint", "location rule", "resource"]

_DEFAULT_DEPENDANT_ELEMENTS = lib.DependantElements({})
_DEFAULT_ELEMENT_REFERENCES = lib.ElementReferences({}, {})
_DEFAULT_UNSUPPORTED_ELEMENTS = lib.UnsupportedElements(
    {}, EXPECTED_TYPES_FOR_REMOVE
)


def _constraints(*argv):
    return f"<constraints>{''.join(argv)}</constraints>"


FIXTURE_LOC_CONSTRAINT_WITH_1_RULE = """
    <rsc_location id="lr1" rsc="A">
        <rule id="r1" boolean-op="and" score="100">
            <expression id="r1e1" operation="eq"
                attribute="#uname" value="node1"/>
            <date_expression id="r1e2" operation="gt"
                start="1970-01-01"/>
        </rule>
    </rsc_location>
"""

FIXTURE_LOC_CONSTRAINT_WITH_2_RULES = """
    <rsc_location id="lr2" rsc="B">
        <rule id="r2" score="-INFINITY" >
            <expression id="r2e1" attribute="pingd"
                operation="lt" value="3000"/>
        </rule>
        <rule id="r3" score-attribute="pingd" >
            <expression id="r3e1" attribute="pingd"
                operation="defined"/>
        </rule>
    </rsc_location>
"""

FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES = _constraints(
    FIXTURE_LOC_CONSTRAINT_WITH_1_RULE,
    FIXTURE_LOC_CONSTRAINT_WITH_2_RULES,
)


def fixture_primitive_to_disable(cib: _Element) -> list[_Element]:
    return [cib.find("./configuration/resources/primitive[@id='A']")]


FIXTURE_GROUP = """
    <group id="G">
        <primitive id="A"/>
        <primitive id="B"/>
    </group>
"""


def fixture_group_to_disable(cib: _Element) -> list[_Element]:
    return [
        cib.find("./configuration/resources/group[@id='G']/primitive[@id='A']"),
        cib.find("./configuration/resources/group[@id='G']/primitive[@id='B']"),
        cib.find("./configuration/resources/group[@id='G']"),
    ]


FIXTURE_CLONE = """
    <clone id="C">
        <primitive id="A"/>
    </clone>
"""


def fixture_clone_to_disable(cib: _Element) -> list[_Element]:
    return [
        cib.find("./configuration/resources/clone[@id='C']/primitive[@id='A']"),
        cib.find("./configuration/resources/clone[@id='C']"),
    ]


class GetCibMixin:
    def get_cib(self, **modifier_shortcuts):
        return etree.fromstring(modify_cib(self.cib, **modifier_shortcuts))


class ElementsToRemoveFindElements(TestCase, GetCibMixin):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.cib = read_test_resource("cib-empty.xml")

    def assert_elements_to_remove(
        self,
        elements_to_remove: lib.ElementsToRemove,
        ids_to_remove: set[str],
        resources_to_remove: Optional[list[etree._Element]] = None,
        dependant_elements: lib.DependantElements = _DEFAULT_DEPENDANT_ELEMENTS,
        element_references: lib.ElementReferences = _DEFAULT_ELEMENT_REFERENCES,
        missing_ids: Optional[set[str]] = None,
        unsupported_elements: lib.UnsupportedElements = _DEFAULT_UNSUPPORTED_ELEMENTS,
    ):
        self.assertEqual(elements_to_remove.ids_to_remove, ids_to_remove)
        self.assertEqual(
            elements_to_remove.resources_to_remove,
            resources_to_remove or [],
        )
        self.assertEqual(
            elements_to_remove.dependant_elements, dependant_elements
        )
        self.assertEqual(
            elements_to_remove.element_references, element_references
        )
        self.assertEqual(elements_to_remove.missing_ids, missing_ids or set())
        self.assertEqual(
            elements_to_remove.unsupported_elements, unsupported_elements
        )

    def test_location_constraint(self):
        cib = self.get_cib(
            constraints="""
                <constraints>
                    <rsc_location id="l1" rsc="A" node="node1" score="200"/>
                    <rsc_location id="l2" rsc="A" node="node2" score="100"/>
                </constraints>
            """
        )

        elements_to_remove = lib.ElementsToRemove(cib, ["l1"])
        self.assert_elements_to_remove(elements_to_remove, {"l1"})

    def test_order_constraint(self):
        cib = self.get_cib(
            constraints="""
                <constraints>
                    <rsc_order id="o1" first="A" then="B"/>
                    <rsc_order id="o2" first="A" then="B"/>
                </constraints>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["o2"])
        self.assert_elements_to_remove(elements_to_remove, {"o2"})

    def test_colocation_constraints(self):
        cib = self.get_cib(
            constraints="""
                <constraints>
                    <rsc_colocation id="c1" rsc="A" with-rsc="B" score="100"/>
                    <rsc_colocation id="c2" rsc="A" with-rsc="B" score="100"/>
                </constraints>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["c1", "c2"])
        self.assert_elements_to_remove(elements_to_remove, {"c1", "c2"})

    def test_ticket_constraints(self):
        cib = self.get_cib(
            constraints="""
                <constraints>
                    <rsc_ticket id="t1" ticket="T" rsc="A"/>
                    <rsc_ticket id="t2" ticket="T" rsc="B"/>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["t1", "t2"])
        self.assert_elements_to_remove(elements_to_remove, {"t1", "t2"})

    def test_location_constraint_with_one_rule_by_id(self):
        cib = self.get_cib(constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES)
        elements_to_remove = lib.ElementsToRemove(cib, ["lr1"])
        self.assert_elements_to_remove(elements_to_remove, {"lr1"})

    def test_location_constraint_with_more_rules_by_id(self):
        cib = self.get_cib(constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES)
        elements_to_remove = lib.ElementsToRemove(cib, ["lr2"])
        self.assert_elements_to_remove(elements_to_remove, {"lr2"})

    def test_one_rule_from_location_constraint_with_one_rule(self):
        cib = self.get_cib(constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES)
        elements_to_remove = lib.ElementsToRemove(cib, ["r1"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"lr1", "r1"},
            dependant_elements=lib.DependantElements(
                {"lr1": const.TAG_CONSTRAINT_LOCATION}
            ),
        )

    def test_one_rule_from_location_constraint_with_two_rules(self):
        cib = self.get_cib(constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES)
        elements_to_remove = lib.ElementsToRemove(cib, ["r2"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"r2"},
            element_references=lib.ElementReferences(
                {"r2": {"lr2"}},
                {
                    "lr2": const.TAG_CONSTRAINT_LOCATION,
                    "r2": const.TAG_RULE,
                },
            ),
        )

    def test_more_location_rules(self):
        cib = self.get_cib(constraints=FIXTURE_TWO_LOC_CONSTRAINTS_WITH_RULES)
        elements_to_remove = lib.ElementsToRemove(cib, ["r1", "r2", "r3"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"r1", "r2", "r3", "lr1", "lr2"},
            dependant_elements=lib.DependantElements(
                {
                    "lr1": const.TAG_CONSTRAINT_LOCATION,
                    "lr2": const.TAG_CONSTRAINT_LOCATION,
                }
            ),
        )

    def test_resource_primitive(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A"},
            resources_to_remove=fixture_primitive_to_disable(cib),
        )

    def test_resource_primitive_in_tag(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """,
            tags="""
                <tags>
                    <tag id="T1">
                        <obj_ref id="A"/>
                        <obj_ref id="B"/>
                    </tag>
                    <tag id="T2">
                        <obj_ref id="A"/>
                    </tag>
                </tags>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "T2"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            dependant_elements=lib.DependantElements({"T2": const.TAG_TAG}),
            element_references=lib.ElementReferences(
                {"A": {"T1"}},
                {"A": const.TAG_RESOURCE_PRIMITIVE, "T1": const.TAG_TAG},
            ),
        )

    def test_resource_primitive_constraints(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="l1" rsc="A" node="node1" score="200"/>
                    <rsc_order id="o1" first="A" then="B"/>
                    <rsc_colocation id="c1" rsc="A" with-rsc="B" score="100"/>
                    <rsc_ticket id="t1" ticket="T" rsc="A"/>
                    <rsc_location id="l2" rsc="B" node="node1" score="200"/>
                </constraints>
            """,
        )

        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "l1", "o1", "c1", "t1"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "l1": const.TAG_CONSTRAINT_LOCATION,
                    "o1": const.TAG_CONSTRAINT_ORDER,
                    "c1": const.TAG_CONSTRAINT_COLOCATION,
                    "t1": const.TAG_CONSTRAINT_TICKET,
                }
            ),
        )

    def test_resource_group(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_GROUP}
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["G"])

        self.assert_elements_to_remove(
            elements_to_remove,
            {"G", "A", "B"},
            resources_to_remove=fixture_group_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "B": const.TAG_RESOURCE_PRIMITIVE,
                }
            ),
        )

    def test_resource_group_member(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_GROUP}
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])

        self.assert_elements_to_remove(
            elements_to_remove,
            {"A"},
            resources_to_remove=[
                cib.find("./configuration/resources/group/primitive[@id='A']"),
            ],
            element_references=lib.ElementReferences(
                {"A": {"G"}},
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "G": const.TAG_RESOURCE_GROUP,
                },
            ),
        )

    def test_resource_group_all_members(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_GROUP}
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A", "B"])

        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "B", "G"},
            resources_to_remove=fixture_group_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {"G": const.TAG_RESOURCE_GROUP}
            ),
        )

    def test_resource_group_constraints(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_GROUP}
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="l1" rsc="G" node="node1" score="200"/>
                    <rsc_order id="o1" first="A" then="B"/>
                    <rsc_colocation id="c1" rsc="A" with-rsc="B" score="100"/>
                    <rsc_ticket id="t1" ticket="T" rsc="B"/>
                    <rsc_location id="l2" rsc="X" node="node1" score="200"/>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["G"])

        self.assert_elements_to_remove(
            elements_to_remove,
            {"G", "A", "B", "l1", "o1", "c1", "t1"},
            resources_to_remove=fixture_group_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "B": const.TAG_RESOURCE_PRIMITIVE,
                    "l1": const.TAG_CONSTRAINT_LOCATION,
                    "o1": const.TAG_CONSTRAINT_ORDER,
                    "c1": const.TAG_CONSTRAINT_COLOCATION,
                    "t1": const.TAG_CONSTRAINT_TICKET,
                }
            ),
        )

    def test_group_in_tag(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_GROUP}
                </resources>
            """,
            tags="""
                <tags>
                    <tag id="T">
                        <obj_ref id="G"/>
                        <obj_ref id="B"/>
                    </tag>
                </tags>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["G"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"G", "A", "B", "T"},
            resources_to_remove=fixture_group_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "B": const.TAG_RESOURCE_PRIMITIVE,
                    "T": const.TAG_TAG,
                }
            ),
        )

    def test_resource_clone(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_CLONE}
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["C"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"C", "A"},
            resources_to_remove=fixture_clone_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {"A": const.TAG_RESOURCE_PRIMITIVE}
            ),
        )

    def test_resource_clone_primitive(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_CLONE}
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])

        self.assert_elements_to_remove(
            elements_to_remove,
            {"C", "A"},
            resources_to_remove=fixture_clone_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {"C": const.TAG_RESOURCE_CLONE}
            ),
        )

    def test_resource_clone_constraints(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_CLONE}
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="l1" rsc="C" node="node1" score="100"/>
                    <rsc_location id="l2" rsc="A" node="node2" score="200"/>
                    <rsc_location id="l3" rsc="X" node="node2" score="200"/>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["C"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"C", "A", "l1", "l2"},
            resources_to_remove=fixture_clone_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "l1": const.TAG_CONSTRAINT_LOCATION,
                    "l2": const.TAG_CONSTRAINT_LOCATION,
                }
            ),
        )

    def test_resource_clone_in_tag(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_CLONE}
                </resources>
            """,
            tags="""
                <tags>
                    <tag id="T">
                        <obj_ref id="C"/>
                        <obj_ref id="A"/>
                    </tag>
                </tags>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["C"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"C", "A", "T"},
            resources_to_remove=fixture_clone_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {"A": const.TAG_RESOURCE_PRIMITIVE, "T": const.TAG_TAG}
            ),
        )

    def test_resource_bundle(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <bundle id="B"/>
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["B"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"B"},
            resources_to_remove=[
                cib.find("./configuration/resources/bundle[@id='B']")
            ],
        )

    def test_resource_bundle_with_primitive(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <bundle id="B">
                        <primitive id="A"/>
                    </bundle>
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["B"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"B", "A"},
            resources_to_remove=[
                cib.find("./configuration/resources/bundle/primitive[@id='A']"),
                cib.find("./configuration/resources/bundle[@id='B']"),
            ],
            dependant_elements=lib.DependantElements(
                {"A": const.TAG_RESOURCE_PRIMITIVE}
            ),
        )

    def test_resource_bundle_primitive(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <bundle id="B">
                        <primitive id="A"/>
                    </bundle>
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A"},
            resources_to_remove=[
                cib.find("./configuration/resources/bundle/primitive[@id='A']")
            ],
            element_references=lib.ElementReferences(
                {"A": {"B"}},
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "B": const.TAG_RESOURCE_BUNDLE,
                },
            ),
        )

    def test_resource_referenced_in_acl(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="vohrablo"/>
                </resources>
            """,
            optional_in_conf="""
                <acls>
                    <acl_role id="zenich">
                        <acl_permission id="ucesat_se1" kind="read" reference="hreben"/>
                        <acl_permission id="ucesat_se2" kind="read" reference="vohrablo"/>
                    </acl_role>
                    <acl_target id="Jirka Kara">
                        <role id="zenich"/>
                    </acl_target>
                </acls>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["vohrablo"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"vohrablo", "ucesat_se2"},
            resources_to_remove=[
                cib.find("./configuration/resources/primitive[@id='vohrablo']")
            ],
            dependant_elements=lib.DependantElements(
                {"ucesat_se2": const.TAG_ACL_PERMISSION}
            ),
            element_references=lib.ElementReferences(
                {"ucesat_se2": {"zenich"}},
                {
                    "ucesat_se2": const.TAG_ACL_PERMISSION,
                    "zenich": const.TAG_ACL_ROLE,
                },
            ),
        )

    def test_resource_referenced_in_acl_indirectly(self):
        cib = self.get_cib(
            resources=f"""
                <resources>
                    {FIXTURE_GROUP}
                </resources>
            """,
            optional_in_conf="""
                <acls>
                    <acl_role id="ROLE">
                        <acl_permission id="PERMISSION" kind="write" reference="A"/>
                    </acl_role>
                    <acl_target id="OtaZnicek">
                        <role id="ROLE"/>
                    </acl_target>
                </acls>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["G"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"G", "A", "B", "PERMISSION"},
            resources_to_remove=fixture_group_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "B": const.TAG_RESOURCE_PRIMITIVE,
                    "PERMISSION": const.TAG_ACL_PERMISSION,
                }
            ),
            element_references=lib.ElementReferences(
                {"PERMISSION": {"ROLE"}},
                {
                    "PERMISSION": const.TAG_ACL_PERMISSION,
                    "ROLE": const.TAG_ACL_ROLE,
                },
            ),
        )

    def test_resource_keep_fencing_level(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """,
            fencing_topology="""
                <fencing-topology>
                    <fencing-level index="2" devices="A,B" target="NODE-A" id="fl"/>
                </fencing-topology>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            element_references=lib.ElementReferences(
                {"A": {"fl"}},
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "fl": const.TAG_FENCING_LEVEL,
                },
            ),
        )

    def test_resource_remove_fencing_level(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                </resources>
            """,
            fencing_topology="""
                <fencing-topology>
                    <fencing-level index="1" devices="A" target="NODE-A" id="fl-NODE-A-1"/>
                </fencing-topology>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "fl-NODE-A-1"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {"fl-NODE-A-1": const.TAG_FENCING_LEVEL}
            ),
        )

    def test_resource_in_constraint_set(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_colocation score="-1" id="c1">
                        <resource_set role="Started" id="set1">
                            <resource_ref id="A"/>
                            <resource_ref id="B"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            element_references=lib.ElementReferences(
                {"A": {"set1"}},
                {
                    "A": const.TAG_RESOURCE_PRIMITIVE,
                    "set1": const.TAG_RESOURCE_SET,
                },
            ),
        )

    def test_resource_in_constraint_set_remove_set_keep_constraint(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <primitive id="B"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_colocation score="-1" id="c1">
                        <resource_set role="Started" id="set1">
                            <resource_ref id="A"/>
                        </resource_set>
                        <resource_set sequential="false" id="set2">
                            <resource_ref id="B"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "set1"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {"set1": const.TAG_RESOURCE_SET}
            ),
            element_references=lib.ElementReferences(
                {"set1": {"c1"}},
                {
                    "set1": const.TAG_RESOURCE_SET,
                    "c1": const.TAG_CONSTRAINT_COLOCATION,
                },
            ),
        )

    def test_resource_in_constraint_set_remove_set_remove_constraint(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_colocation score="-1" id="c1">
                        <resource_set role="Started" id="set1">
                            <resource_ref id="A"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "set1", "c1"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "set1": const.TAG_RESOURCE_SET,
                    "c1": const.TAG_CONSTRAINT_COLOCATION,
                }
            ),
        )

    def test_resource_in_multiple_sets(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_colocation score="-1" id="c1">
                        <resource_set role="Started" id="set1">
                            <resource_ref id="A"/>
                        </resource_set>
                    </rsc_colocation>
                    <rsc_colocation score="1" id="c2">
                        <resource_set role="Started" id="set2">
                            <resource_ref id="A"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "set1", "c1", "set2", "c2"},
            resources_to_remove=fixture_primitive_to_disable(cib),
            dependant_elements=lib.DependantElements(
                {
                    "set1": const.TAG_RESOURCE_SET,
                    "c1": const.TAG_CONSTRAINT_COLOCATION,
                    "set2": const.TAG_RESOURCE_SET,
                    "c2": const.TAG_CONSTRAINT_COLOCATION,
                }
            ),
        )

    def test_resource_legacy_promotable_clone(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <master id="MS">
                        <primitive id="A"/>
                    </master>
                </resources>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["MS"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"MS", "A"},
            resources_to_remove=[
                cib.find("./configuration/resources/master/primitive[@id='A']"),
                cib.find("./configuration/resources/master/[@id='MS']"),
            ],
            dependant_elements=lib.DependantElements(
                {"A": const.TAG_RESOURCE_PRIMITIVE}
            ),
        )

    def test_resource_legacy_promotable_clone_inner_element(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <master id="MS">
                        <primitive id="A"/>
                    </master>
                </resources>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A", "MS"},
            resources_to_remove=[
                cib.find("./configuration/resources/master/primitive[@id='A']"),
                cib.find("./configuration/resources/master/[@id='MS']"),
            ],
            dependant_elements=lib.DependantElements({"MS": "master"}),
        )

    def test_missing_elements(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                </resources>
            """
        )
        elements_to_remove = lib.ElementsToRemove(cib, ["A", "B", "C", "D"])
        self.assert_elements_to_remove(
            elements_to_remove,
            {"A"},
            missing_ids={"B", "C", "D"},
            resources_to_remove=fixture_primitive_to_disable(cib),
        )

    def test_unsupported_id_types(self):
        cib = self.get_cib(
            tags="""
                <tags>
                    <tag id="T">
                        <obj_ref id="A"/>
                    </tag>
                </tags>
            """,
            fencing_topology="""
                <fencing-topology>
                    <fencing-level index="1" devices="A" target="NODE-A" id="fl"/>
                </fencing-topology>
            """,
            optional_in_conf="""
                <acls>
                    <acl_role id="role">
                        <acl_permission id="role-read" kind="read" reference="A"/>
                    </acl_role>
                </acls>
            """,
        )
        elements_to_remove = lib.ElementsToRemove(
            cib, ["T", "fl", "role", "role-read"]
        )
        self.assert_elements_to_remove(
            elements_to_remove,
            set(),
            unsupported_elements=lib.UnsupportedElements(
                {
                    "T": const.TAG_TAG,
                    "fl": const.TAG_FENCING_LEVEL,
                    "role": const.TAG_ACL_ROLE,
                    "role-read": const.TAG_ACL_PERMISSION,
                },
                EXPECTED_TYPES_FOR_REMOVE,
            ),
        )


class RemoveSpecifiedElements(TestCase, GetCibMixin):
    def setUp(self):
        self.elements_to_remove_mock = mock.Mock()
        self.elements_to_remove_mock.ids_to_remove = set()
        self.elements_to_remove_mock.element_references = lib.ElementReferences(
            {}, {}
        )
        self.cib = read_test_resource("cib-empty.xml")

    def test_remove_nothing_to_remove(self):
        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <clone id="C">
                        <primitive id="C-1"/>
                    </clone>
                    <primitive id="X"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="B"/>
                    <rsc_location id="Y"/>
                </constraints>
            """,
            tags="""
                <tags>
                    <tag id="D"/>
                    <tag id="Z"/>
                </tags>
            """,
        )

        initial_cib = etree_to_str(cib)

        lib.remove_specified_elements(cib, self.elements_to_remove_mock)

        assert_xml_equal(initial_cib, etree_to_str(cib))

    def test_remove_only_by_id(self):
        self.elements_to_remove_mock.ids_to_remove = {"A", "B", "C", "D"}

        cib = self.get_cib(
            resources="""
                <resources>
                    <primitive id="A"/>
                    <clone id="C">
                        <primitive id="C-1"/>
                    </clone>
                    <primitive id="X"/>
                </resources>
            """,
            constraints="""
                <constraints>
                    <rsc_location id="B"/>
                    <rsc_location id="Y"/>
                </constraints>
            """,
            tags="""
                <tags>
                    <tag id="D"/>
                    <tag id="Z"/>
                </tags>
            """,
        )

        lib.remove_specified_elements(cib, self.elements_to_remove_mock)

        assert_xml_equal(
            modify_cib(
                self.cib,
                resources="""
                    <resources>
                        <primitive id="X"/>
                    </resources>
                """,
                constraints="""
                    <constraints>
                        <rsc_location id="Y"/>
                    </constraints>
                """,
                tags="""
                    <tags>
                        <tag id="Z"/>
                    </tags>
                """,
            ),
            etree_to_str(cib),
        )

    def test_remove_reference_from_tag(self):
        self.elements_to_remove_mock.element_references = lib.ElementReferences(
            {"A": {"T1"}},
            {"A": const.TAG_RESOURCE_PRIMITIVE, "T1": const.TAG_TAG},
        )

        cib = self.get_cib(
            tags="""
                <tags>
                    <tag id="T1">
                        <obj_ref id="A"/>
                        <obj_ref id="B"/>
                    </tag>
                    <tag id="T2">
                        <obj_ref id="A"/>
                        <obj_ref id="B"/>
                    </tag>
                </tags>
            """,
        )

        lib.remove_specified_elements(cib, self.elements_to_remove_mock)

        assert_xml_equal(
            modify_cib(
                self.cib,
                tags="""
                    <tags>
                        <tag id="T1">
                            <obj_ref id="B"/>
                        </tag>
                        <tag id="T2">
                            <obj_ref id="A"/>
                            <obj_ref id="B"/>
                        </tag>
                    </tags>
                """,
            ),
            etree_to_str(cib),
        )

    def test_remove_reference_from_resource_set(self):
        self.elements_to_remove_mock.element_references = lib.ElementReferences(
            {"A": {"set1"}},
            {
                "A": const.TAG_RESOURCE_PRIMITIVE,
                "set1": const.TAG_RESOURCE_SET,
            },
        )

        cib = self.get_cib(
            constraints="""
                <constraints>
                    <rsc_colocation score="-1" id="c1">
                        <resource_set role="Started" id="set1">
                            <resource_ref id="A"/>
                            <resource_ref id="B"/>
                        </resource_set>
                    </rsc_colocation>
                    <rsc_colocation score="1" id="c2">
                        <resource_set role="Started" id="set2">
                            <resource_ref id="A"/>
                            <resource_ref id="B"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """,
        )

        lib.remove_specified_elements(cib, self.elements_to_remove_mock)

        assert_xml_equal(
            modify_cib(
                self.cib,
                constraints="""
                    <constraints>
                        <rsc_colocation score="-1" id="c1">
                            <resource_set role="Started" id="set1">
                                <resource_ref id="B"/>
                            </resource_set>
                        </rsc_colocation>
                        <rsc_colocation score="1" id="c2">
                            <resource_set role="Started" id="set2">
                                <resource_ref id="A"/>
                                <resource_ref id="B"/>
                            </resource_set>
                        </rsc_colocation>
                    </constraints>
                """,
            ),
            etree_to_str(cib),
        )

    def test_remove_references_from_fencing_level(self):
        self.elements_to_remove_mock.element_references = lib.ElementReferences(
            {"A": {"FL1"}},
            {"A": const.TAG_RESOURCE_PRIMITIVE, "FL1": const.TAG_FENCING_LEVEL},
        )

        cib = self.get_cib(
            fencing_topology="""
                <fencing-topology>
                    <fencing-level index="1" devices="A,B" target="NODE-A" id="FL1"/>
                    <fencing-level index="2" devices="A,B" target="NODE-A" id="FL2"/>
                </fencing-topology>
            """,
        )

        lib.remove_specified_elements(cib, self.elements_to_remove_mock)

        assert_xml_equal(
            modify_cib(
                self.cib,
                fencing_topology="""
                    <fencing-topology>
                        <fencing-level index="1" devices="B" target="NODE-A" id="FL1"/>
                        <fencing-level index="2" devices="A,B" target="NODE-A" id="FL2"/>
                    </fencing-topology>
                """,
            ),
            etree_to_str(cib),
        )

    def test_remove_references_from_ignored_type(self):
        self.elements_to_remove_mock.element_references = lib.ElementReferences(
            {"A": {"G"}},
            {"A": const.TAG_RESOURCE_PRIMITIVE, "G": const.TAG_RESOURCE_GROUP},
        )

        cib = self.get_cib(
            resources="""
                <resources>
                    <group id="G">
                        <primitive id="A"/>
                    </group>
                </resources>
            """,
        )

        initial_cib = etree_to_str(cib)

        lib.remove_specified_elements(cib, self.elements_to_remove_mock)

        assert_xml_equal(initial_cib, etree_to_str(cib))


class EnsureResourcesStopped(TestCase):
    def setUp(self):
        self.state_xml = read_test_resource("crm_mon.minimal.xml")

    def test_ok(self):
        state = complete_state(
            self.state_xml,
            """
                <resources>
                    <resource id="A" managed="true" role="Stopped"/>
                    <group id="B" number_resources="1" maintenance="false" managed="true" disabled="false">
                        <resource id="1" role="Stopped" managed="true"/>
                    </group>
                    <clone id="C" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <resource id="2" role="Stopped" managed="true"/>
                    </clone>
                </resources>
            """,
        )

        report_list = lib.ensure_resources_stopped(state, ["A", "B", "C"], [])
        self.assertEqual(report_list, [])

    def test_some_not_stopped(self):
        state = complete_state(
            self.state_xml,
            """
                <resources>
                    <resource id="A" managed="true" role="Stopped"/>
                    <group id="B" number_resources="1" maintenance="false" managed="true" disabled="false">
                        <resource id="1" role="Started" managed="true"/>
                    </group>
                    <clone id="C" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <resource id="2" role="Stopped" managed="true"/>
                    </clone>
                </resources>
            """,
        )

        report_list = lib.ensure_resources_stopped(state, ["A", "B", "C"], [])
        self.assertEqual(
            report_list,
            [
                reports.ReportItem.error(
                    reports.messages.CannotRemoveResourcesNotStopped(["B"]),
                    force_code=reports.codes.FORCE,
                )
            ],
        )

    def test_not_stopped_forced(self):
        state = complete_state(
            self.state_xml,
            """
                <resources>
                    <resource id="A" managed="true" role="Started"/>
                </resources>
            """,
        )

        report_list = lib.ensure_resources_stopped(
            state, ["A"], [reports.codes.FORCE]
        )
        self.assertEqual(
            report_list,
            [
                reports.ReportItem.warning(
                    reports.messages.CannotRemoveResourcesNotStopped(["A"])
                )
            ],
        )

    def test_multiinstance_some_not_stopped_clone_id(self):
        state = complete_state(
            self.state_xml,
            """
                <resources>
                    <clone id="C" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <resource id="2" role="Stopped" managed="true"/>
                        <resource id="2" role="Started" managed="true"/>
                    </clone>
                </resources>
            """,
        )

        report_list = lib.ensure_resources_stopped(state, ["C"], [])
        self.assertEqual(
            report_list,
            [
                reports.ReportItem.error(
                    reports.messages.CannotRemoveResourcesNotStopped(["C"]),
                    force_code=reports.codes.FORCE,
                )
            ],
        )

    def test_multiinstance_some_not_stopped_primitive_id(self):
        state = complete_state(
            self.state_xml,
            """
                <resources>
                    <clone id="C" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                        <resource id="A" role="Stopped" managed="true"/>
                        <resource id="A" role="Started" managed="true"/>
                    </clone>
                </resources>
            """,
        )

        report_list = lib.ensure_resources_stopped(state, ["A"], [])
        self.assertEqual(
            report_list,
            [
                reports.ReportItem.error(
                    reports.messages.CannotRemoveResourcesNotStopped(["A"]),
                    force_code=reports.codes.FORCE,
                )
            ],
        )

    def test_works_with_clones_and_bundle_in_status(self):
        state = complete_state(
            self.state_xml,
            """
            <resources>
                <bundle id="BUNDLE" type="podman" image="localhost/pcmktest:http" unique="false" maintenance="false" managed="true" failed="false">
                    <replica id="0">
                        <resource id="BUNDLE-ip-192.168.122.250" resource_agent="ocf:heartbeat:IPaddr2" role="Stopped"/>
                        <resource id="A" role="Stopped"/>
                        <resource id="BUNDLE-podman-0" resource_agent="ocf:heartbeat:podman" role="Stopped"/>
                        <resource id="BUNDLE-0" resource_agent="ocf:pacemaker:remote" role="Stopped"/>
                    </replica>
                </bundle>
                <clone id="C" multi_state="false" unique="false" maintenance="false" managed="true" disabled="false" failed="false" failure_ignored="false">
                    <resource id="2" role="Stopped" managed="true"/>
                </clone>
            </resources>
        """,
        )

        report_list = lib.ensure_resources_stopped(state, ["C"], [])
        self.assertEqual(report_list, [])


class DependantElementsToReports(TestCase):
    # pylint: disable=no-self-use
    def test_no_reports(self):
        elements = lib.DependantElements({})
        assert_report_item_list_equal(elements.to_reports(), [])

    def test_reports(self):
        elements = lib.DependantElements(
            {"A": const.TAG_RESOURCE_PRIMITIVE, "B": const.TAG_TAG}
        )
        assert_report_item_list_equal(
            elements.to_reports(),
            [
                fixture.info(
                    reports.codes.CIB_REMOVE_DEPENDANT_ELEMENTS,
                    id_tag_map={
                        "A": const.TAG_RESOURCE_PRIMITIVE,
                        "B": const.TAG_TAG,
                    },
                )
            ],
        )


class ElementReferencesToReports(TestCase):
    # pylint: disable=no-self-use
    def test_no_reports(self):
        elements = lib.ElementReferences({}, {})
        assert_report_item_list_equal(elements.to_reports(), [])

    def test_reports(self):
        elements = lib.ElementReferences(
            {"A": {"B"}, "C": {"B"}},
            {
                "A": const.TAG_RESOURCE_PRIMITIVE,
                "B": const.TAG_TAG,
                "C": const.TAG_RESOURCE_PRIMITIVE,
            },
        )
        assert_report_item_list_equal(
            elements.to_reports(),
            [
                fixture.info(
                    reports.codes.CIB_REMOVE_REFERENCES,
                    removing_references_from={"A": {"B"}, "C": {"B"}},
                    id_tag_map={
                        "A": const.TAG_RESOURCE_PRIMITIVE,
                        "B": const.TAG_TAG,
                        "C": const.TAG_RESOURCE_PRIMITIVE,
                    },
                )
            ],
        )


class GetInnerReferences(TestCase):
    # pylint: disable=protected-access
    def test_no_inner_references(self):
        self.assertEqual(
            [], lib._get_inner_references(etree.fromstring("<A/>"))
        )

    def test_not_supported_inner_references(self):
        element = etree.fromstring(
            """
            <A>
                <a id="1">
                    <a id="11"/>
                    <a id="12"/>
                </a>
                <b id="2"/>
                <c/>
            </A>
            """
        )
        self.assertEqual([], lib._get_inner_references(element))

    def test_resource_primitive(self):
        primitive = etree.fromstring('<primitive id="A"/>')
        self.assertEqual([], lib._get_inner_references(primitive))

    def test_resource_group(self):
        element = etree.fromstring('<group id="G"/>')
        child1 = etree.SubElement(element, "primitive", id="A")
        child2 = etree.SubElement(element, "primitive", id="B")

        self.assertEqual([child1, child2], lib._get_inner_references(element))

    def test_resource_clone(self):
        element = etree.fromstring('<clone id="C"/>')
        child = etree.SubElement(element, "primitive", id="A")

        self.assertEqual([child], lib._get_inner_references(element))

    def test_resource_bundle(self):
        element = etree.fromstring('<bundle id="B"/>')
        child = etree.SubElement(element, "primitive", id="A")

        self.assertEqual([child], lib._get_inner_references(element))

    def test_resource_bundle_no_primitive(self):
        element = etree.fromstring('<bundle id="B"/>')

        self.assertEqual([], lib._get_inner_references(element))


class IsLastElement(TestCase):
    # pylint: disable=protected-access
    def test_last_element_true(self):
        for element, tag in (
            (etree.fromstring("<A><a/></A>"), "a"),
            (etree.fromstring("<A><a/><b/><c/></A>"), "a"),
        ):
            with self.subTest(element=element, tag=tag):
                self.assertTrue(lib._is_last_element(element, tag))

    def test_last_element_false(self):
        for element, tag in (
            (etree.fromstring("<A/>"), "a"),
            (etree.fromstring("<A> <a/> <a/> </A>"), "a"),
            (etree.fromstring("<A><A> <a/> </A></A>"), "a"),
        ):
            with self.subTest(element=element, tag=tag):
                self.assertFalse(lib._is_last_element(element, tag))


class IsEmptyAfterInnerElRemoval(TestCase):
    # pylint: disable=protected-access
    def test_last_element_true(self):
        for parent in (
            etree.fromstring("<clone/>"),
            etree.fromstring("<clone><primitive/></clone>"),
            etree.fromstring("<group><primitive/></group>"),
            etree.fromstring("<tag><obj_ref/></tag>"),
            etree.fromstring("<resource_set><resource_ref/></resource_set>"),
            etree.fromstring("<rsc_order><resource_set/></rsc_order>"),
            etree.fromstring("<rsc_location><rule/></rsc_location>"),
        ):
            with self.subTest(parent=parent.tag):
                self.assertTrue(lib._is_empty_after_inner_el_removal(parent))

    def test_last_element_false(self):
        for parent in (
            etree.fromstring("<bundle/>"),
            etree.fromstring("<bundle><primitive/></bundle>"),
            etree.fromstring("<primitive/>"),
            etree.fromstring("<group/>"),
            etree.fromstring("<group><primitive/><primitive/></group>"),
            etree.fromstring("<tag/>"),
            etree.fromstring("<tag><obj_ref/><obj_ref/></tag>"),
            etree.fromstring(
                """
                <resource_set>
                    <resource_ref/>
                    <resource_ref/>
                </resource_set>
               """
            ),
            etree.fromstring("<resource_set/>"),
            etree.fromstring(
                "<element><resource_set/><resource_set/></element>"
            ),
            etree.fromstring(
                """
                <element>
                    <resource_set/>
                    <resource_set/>"
                </element>
                """
            ),
            etree.fromstring("<rsc_location><rule/><rule/></rsc_location>"),
            etree.fromstring("<rsc_location/>"),
        ):
            with self.subTest(parent=parent):
                self.assertFalse(lib._is_empty_after_inner_el_removal(parent))



from typing import (
    Any,
    Optional,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    FUTURE_OPTION,
    Argv,
    InputModifiers,
    KeyValueParser,
    group_by_keywords,
)
from pcs.cli.reports.output import (
    deprecation_warning,
    process_library_reports,
    warn,
)
from pcs.common import reports
from pcs.common.reports import codes as report_codes
from pcs.lib.errors import LibraryError


def config_setup(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    create booth config

    Options:
      * --force - overwrite existing
      * --booth-conf - booth config file
      * --booth-key - booth authkey file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported(
        "--force",
        "--booth-conf",
        "--booth-key",
        "--name",
    )
    peers = group_by_keywords(arg_list, {"sites", "arbitrators"})
    peers.ensure_unique_keywords()
    if not peers.has_keyword("sites") or not peers.get_args_flat("sites"):
        raise CmdLineInputError()

    lib.booth.config_setup(
        peers.get_args_flat("sites"),
        peers.get_args_flat("arbitrators"),
        instance_name=modifiers.get("--name"),
        overwrite_existing=modifiers.get("--force"),
    )


def config_destroy(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    destroy booth config

    Options:
      --force - ignore config load issues
      --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--force", "--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.config_destroy(
        instance_name=modifiers.get("--name"),
        ignore_config_load_problems=modifiers.get("--force"),
    )


def config_show(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    print booth config

    Options:
      * --name - name of a booth instance
      * --request-timeout - HTTP timeout for getting config from remote host
    """
    modifiers.ensure_only_supported("--name", "--request-timeout")
    if len(arg_list) > 1:
        raise CmdLineInputError()
    node = None if not arg_list else arg_list[0]

    print(
        lib.booth.config_text(
            instance_name=modifiers.get("--name"), node_name=node
        )
        .decode("utf-8")
        .rstrip()
    )


def config_ticket_add(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    add ticket to current configuration

    Options:
      * --force
      * --booth-conf - booth config file
      * --booth-key - booth auth key file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported(
        "--force", "--booth-conf", "--name", "--booth-key"
    )
    if not arg_list:
        raise CmdLineInputError
    lib.booth.config_ticket_add(
        arg_list[0],
        KeyValueParser(arg_list[1:]).get_unique(),
        instance_name=modifiers.get("--name"),
        allow_unknown_options=modifiers.get("--force"),
    )


def config_ticket_remove(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    add ticket to current configuration

    Options:
      * --booth-conf - booth config file
      * --booth-key - booth auth key file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--booth-conf", "--name", "--booth-key")
    if len(arg_list) != 1:
        raise CmdLineInputError
    lib.booth.config_ticket_remove(
        arg_list[0],
        instance_name=modifiers.get("--name"),
    )


def enable_authfile(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --booth-conf - booth config file
      * --booth-key - booth auth key file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--booth-conf", "--name", "--booth-key")
    if len(arg_list):
        raise CmdLineInputError()
    lib.booth.config_set_enable_authfile(instance_name=modifiers.get("--name"))


def enable_authfile_clean(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --booth-conf - booth config file
      * --booth-key - booth auth key file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--booth-conf", "--name", "--booth-key")
    if len(arg_list):
        raise CmdLineInputError()
    lib.booth.config_unset_enable_authfile(
        instance_name=modifiers.get("--name")
    )


def _parse_ticket_operation(arg_list: Argv) -> tuple[str, Optional[str]]:
    site_ip = None
    if len(arg_list) == 2:
        site_ip = arg_list[1]
    elif len(arg_list) != 1:
        raise CmdLineInputError()
    ticket = arg_list[0]
    return ticket, site_ip


def ticket_revoke(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    ticket, site_ip = _parse_ticket_operation(arg_list)
    lib.booth.ticket_revoke(
        ticket, site_ip=site_ip, instance_name=modifiers.get("--name")
    )


def ticket_grant(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    ticket, site_ip = _parse_ticket_operation(arg_list)
    lib.booth.ticket_grant(
        ticket, site_ip=site_ip, instance_name=modifiers.get("--name")
    )


def ticket_cleanup(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    if not arg_list:
        modifiers.ensure_only_supported("--name")
        lib.booth.ticket_cleanup_auto(instance_name=modifiers.get("--name"))
        return

    if len(arg_list) != 1:
        raise CmdLineInputError()
    modifiers.ensure_only_supported()
    lib.booth.ticket_cleanup(arg_list[0])


def ticket_unstandby(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: None
    """
    modifiers.ensure_only_supported()
    if len(arg_list) != 1:
        raise CmdLineInputError()
    lib.booth.ticket_unstandby(arg_list[0])


def ticket_standby(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options: None
    """
    modifiers.ensure_only_supported()
    if len(arg_list) != 1:
        raise CmdLineInputError()
    lib.booth.ticket_standby(arg_list[0])


def create_in_cluster(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allows to create booth resource even if its agent is not
        installed
      * -f - CIB file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--force", "-f", "--name")
    if len(arg_list) != 2 or arg_list[0] != "ip":
        raise CmdLineInputError()
    lib.booth.create_in_cluster(
        arg_list[1],
        instance_name=modifiers.get("--name"),
        allow_absent_resource_agent=modifiers.get("--force"),
    )


def remove_from_cluster(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow remove of multiple, (deprecated) don't stop resources
      * -f - CIB file
      * --name - name of a booth instance
      * --no-stop - don't stop resources before deletion
      * --future - specifying '--force' does not skip resource stopping
    """

    def _process_reports(
        report_list: reports.ReportItemList,
        force_flags: reports.types.ForceFlags,
    ) -> reports.ReportItemList:
        filtered_reports = []
        for report in report_list:
            if (
                report.message.code
                == reports.codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED
            ):
                continue
            if (
                report.severity.level == reports.ReportItemSeverity.ERROR
                and report.severity.force_code in force_flags
            ):
                report.severity = reports.ReportItemSeverity.warning()
            filtered_reports.append(report)
        return filtered_reports

    modifiers.ensure_only_supported(
        "-f", "--force", FUTURE_OPTION, "--name", "--no-stop"
    )
    modifiers.ensure_not_mutually_exclusive("-f", "--no-stop")
    if arg_list:
        raise CmdLineInputError()

    force_flags = []
    if modifiers.get("--force"):
        force_flags.append(report_codes.FORCE)

    instance_name = modifiers.get("--name")

    if modifiers.is_specified("-f"):
        warn(
            "Resources are not going to be stopped before deletion because the "
            "command does not run on a live cluster"
        )
        lib.booth.remove_from_cluster(instance_name, force_flags)
        return

    dont_stop_me_now = modifiers.is_specified("--no-stop")
    if (
        not modifiers.is_specified(FUTURE_OPTION)
        and modifiers.is_specified("--force")
        and not dont_stop_me_now
    ):
        # deprecated after pcs-0.12.0
        deprecation_warning(
            "Using '--force' to skip resource stopping is deprecated and will "
            "be removed in a future release. Specify '--future' to switch to "
            "the future behavior and use '--no-stop' to skip resource stopping."
        )
        dont_stop_me_now = True

    if dont_stop_me_now:
        lib.booth.remove_from_cluster(instance_name, force_flags)
        return

    original_report_processor = lib.env.report_processor
    in_memory_report_processor = reports.processor.ReportProcessorInMemory()
    lib.env.report_processor = in_memory_report_processor

    try:
        lib.booth.remove_from_cluster(instance_name)

        if in_memory_report_processor.reports:
            process_library_reports(
                in_memory_report_processor.reports,
                include_debug=modifiers.is_specified("--debug"),
            )
        return
    except LibraryError as e:
        filtered_reports = _process_reports(
            in_memory_report_processor.reports, force_flags
        )

        if reports.has_errors(filtered_reports) or e.output or e.args:
            if filtered_reports:
                process_library_reports(
                    filtered_reports,
                    include_debug=modifiers.is_specified("--debug"),
                    exit_on_error=False,
                )

            raise e

    lib.env.report_processor = original_report_processor

    resource_ids = lib.booth.get_resource_ids_from_cluster(instance_name)
    lib.resource.stop(resource_ids, force_flags)
    lib.cluster.wait_for_pcmk_idle(None)

    lib.booth.remove_from_cluster(instance_name, force_flags)


def restart(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - allow multiple
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--force", "--name")
    if arg_list:
        raise CmdLineInputError()

    lib.booth.restart(
        instance_name=modifiers.get("--name"),
        allow_multiple=modifiers.get("--force"),
    )


def sync(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --skip-offline - skip offline nodes
      * --name - name of a booth instance
      * --booth-conf - booth config file
      * --booth-key - booth authkey file
      * --request-timeout - HTTP timeout for file distribution
    """
    modifiers.ensure_only_supported(
        "--skip-offline",
        "--name",
        "--booth-conf",
        "--booth-key",
        "--request-timeout",
    )
    if arg_list:
        raise CmdLineInputError()
    lib.booth.config_sync(
        instance_name=modifiers.get("--name"),
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )


def enable(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.enable_booth(instance_name=modifiers.get("--name"))


def disable(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.disable_booth(instance_name=modifiers.get("--name"))


def start(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.start_booth(instance_name=modifiers.get("--name"))


def stop(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.stop_booth(instance_name=modifiers.get("--name"))


def pull(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of a booth instance
      * --request-timeout - HTTP timeout for file distribution
    """
    modifiers.ensure_only_supported("--name", "--request-timeout")
    if len(arg_list) != 1:
        raise CmdLineInputError()
    lib.booth.pull_config(
        arg_list[0],
        instance_name=modifiers.get("--name"),
    )


def status(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --name - name of booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    booth_status = lib.booth.get_status(instance_name=modifiers.get("--name"))
    if booth_status.get("ticket"):
        print("TICKETS:")
        print(booth_status["ticket"])
    if booth_status.get("peers"):
        print("PEERS:")
        print(booth_status["peers"])
    if booth_status.get("status"):
        print("DAEMON STATUS:")
        print(booth_status["status"])



from typing import NewType

from pcs.common.tools import Version

PcmkRoleType = NewType("PcmkRoleType", str)
PcmkStatusRoleType = NewType("PcmkStatusRoleType", str)
PcmkOnFailAction = NewType("PcmkOnFailAction", str)
PcmkAction = NewType("PcmkAction", str)
ResourceIdType = NewType("ResourceIdType", str)

INFINITY = "INFINITY"
PCMK_ROLE_STARTED = PcmkRoleType("Started")
PCMK_ROLE_STOPPED = PcmkRoleType("Stopped")
PCMK_ROLE_PROMOTED = PcmkRoleType("Promoted")
PCMK_ROLE_UNPROMOTED = PcmkRoleType("Unpromoted")
PCMK_ROLE_PROMOTED_LEGACY = PcmkRoleType("Master")
PCMK_ROLE_UNPROMOTED_LEGACY = PcmkRoleType("Slave")
PCMK_STATUS_ROLE_STARTED = PcmkStatusRoleType("Started")
PCMK_STATUS_ROLE_STOPPED = PcmkStatusRoleType("Stopped")
PCMK_STATUS_ROLE_PROMOTED = PcmkStatusRoleType("Promoted")
PCMK_STATUS_ROLE_UNPROMOTED = PcmkStatusRoleType("Unpromoted")
PCMK_STATUS_ROLE_STARTING = PcmkStatusRoleType("Starting")
PCMK_STATUS_ROLE_STOPPING = PcmkStatusRoleType("Stopping")
PCMK_STATUS_ROLE_MIGRATING = PcmkStatusRoleType("Migrating")
PCMK_STATUS_ROLE_PROMOTING = PcmkStatusRoleType("Promoting")
PCMK_STATUS_ROLE_DEMOTING = PcmkStatusRoleType("Demoting")
PCMK_ON_FAIL_ACTION_IGNORE = PcmkOnFailAction("ignore")
PCMK_ON_FAIL_ACTION_BLOCK = PcmkOnFailAction("block")
PCMK_ON_FAIL_ACTION_DEMOTE = PcmkOnFailAction("demote")
PCMK_ON_FAIL_ACTION_STOP = PcmkOnFailAction("stop")
PCMK_ON_FAIL_ACTION_RESTART = PcmkOnFailAction("restart")
PCMK_ON_FAIL_ACTION_STANDBY = PcmkOnFailAction("standby")
PCMK_ON_FAIL_ACTION_FENCE = PcmkOnFailAction("fence")
PCMK_ON_FAIL_ACTION_RESTART_CONTAINER = PcmkOnFailAction("restart-container")
PCMK_ROLES_RUNNING = (
    PCMK_ROLE_STARTED,
    PCMK_ROLE_PROMOTED,
    PCMK_ROLE_UNPROMOTED,
)
PCMK_ROLES_RUNNING_WITH_LEGACY = PCMK_ROLES_RUNNING + (
    PCMK_ROLE_PROMOTED_LEGACY,
    PCMK_ROLE_UNPROMOTED_LEGACY,
)
PCMK_ROLES = (PCMK_ROLE_STOPPED,) + PCMK_ROLES_RUNNING
PCMK_ROLES_WITH_LEGACY = (PCMK_ROLE_STOPPED,) + PCMK_ROLES_RUNNING_WITH_LEGACY
PCMK_STATUS_ROLES_RUNNING = (
    PCMK_STATUS_ROLE_STARTED,
    PCMK_STATUS_ROLE_PROMOTED,
    PCMK_STATUS_ROLE_UNPROMOTED,
)
PCMK_STATUS_ROLES_PENDING = (
    PCMK_STATUS_ROLE_STARTING,
    PCMK_STATUS_ROLE_STOPPING,
    PCMK_STATUS_ROLE_MIGRATING,
    PCMK_STATUS_ROLE_PROMOTING,
    PCMK_STATUS_ROLE_DEMOTING,
)
PCMK_STATUS_ROLES = (
    PCMK_STATUS_ROLES_RUNNING
    + PCMK_STATUS_ROLES_PENDING
    + (PCMK_STATUS_ROLE_STOPPED,)
)
PCMK_ACTION_START = PcmkAction("start")
PCMK_ACTION_STOP = PcmkAction("stop")
PCMK_ACTION_PROMOTE = PcmkAction("promote")
PCMK_ACTION_DEMOTE = PcmkAction("demote")
PCMK_ACTIONS = (
    PCMK_ACTION_START,
    PCMK_ACTION_STOP,
    PCMK_ACTION_PROMOTE,
    PCMK_ACTION_DEMOTE,
)
PCMK_NEW_ROLES_CIB_VERSION = Version(3, 7, 0)
# CIB schema which supports new rules syntax and options defined in Pacemaker 3.
# Lower schema is not supported for rules by pcs, as we support Pacemaker 3+
# only in pcs-0.12.
PCMK_RULES_PCMK3_SYNTAX_CIB_VERSION = Version(3, 9, 0)
PCMK_ON_FAIL_DEMOTE_CIB_VERSION = Version(3, 4, 0)

RESOURCE_ID_TYPE_PLAIN = ResourceIdType("resource_id_plain")
RESOURCE_ID_TYPE_REGEXP = ResourceIdType("resource_id_regexp")



from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common.const import PcmkRoleType
from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.types import CibResourceDiscovery

from ..rule import CibRuleExpressionDto
from .set import CibResourceSetDto


@dataclass(frozen=True)
class CibConstraintLocationAttributesDto(DataTransferObject):
    constraint_id: str
    score: Optional[str]
    node: Optional[str]
    rules: Sequence[CibRuleExpressionDto]
    lifetime: Sequence[CibRuleExpressionDto]
    resource_discovery: Optional[CibResourceDiscovery]


@dataclass(frozen=True)
class CibConstraintLocationDto(DataTransferObject):
    resource_id: Optional[str]
    resource_pattern: Optional[str]
    role: Optional[PcmkRoleType]
    attributes: CibConstraintLocationAttributesDto


@dataclass(frozen=True)
class CibConstraintLocationSetDto(DataTransferObject):
    resource_sets: Sequence[CibResourceSetDto]
    attributes: CibConstraintLocationAttributesDto



from pcs.lib.resource_agent import const as ra_const

from pcs_test.tier1.cib_resource.common import (
    ResourceTest,
    fixture_meta_attributes_warning,
)
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.misc import is_minimum_pacemaker_version

PCMK_2_0_3_PLUS = is_minimum_pacemaker_version(2, 0, 3)
PCMK_2_0_5_PLUS = is_minimum_pacemaker_version(2, 0, 5)
ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)


class PlainStonith(ResourceTest):
    def setUp(self):
        super().setUp()
        self.pcs_runner.mock_settings = get_mock_settings()

    def test_simplest(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_base_with_agent_that_provides_unfencing(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_unfencing".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_unfencing">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-provides" name="provides"
                            value="unfencing"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_error_when_not_valid_name(self):
        self.assert_pcs_fail_regardless_of_force(
            "stonith create S fence_pcsmock:invalid".split(),
            "Error: Invalid stonith agent name 'fence_pcsmock:invalid'. Agent name "
            "cannot contain the ':' character, do not use the 'stonith:' prefix. "
            "List of agents can be obtained by using command 'pcs stonith list'.\n"
            + ERRORS_HAVE_OCCURRED,
        )

    def test_error_when_not_valid_agent(self):
        self.assert_pcs_fail(
            "stonith create S absent".split(),
            stderr_full=(
                "Error: Agent 'stonith:absent' is not installed or "
                "does not provide valid metadata: "
                "pcs mock error message: unable to load agent metadata, "
                "use --force to override\n" + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_warning_when_not_valid_agent(self):
        self.assert_effect(
            "stonith create S absent --force".split(),
            """<resources>
                <primitive class="stonith" id="S" type="absent">
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            stderr_full=(
                "Warning: Agent 'stonith:absent' is not installed or "
                "does not provide valid metadata: "
                "pcs mock error message: unable to load agent metadata\n"
            ),
        )

    def test_disabled_puts_target_role_stopped(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal --disabled".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_debug_and_verbose_allowed(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_params ip=i username=u verbose=v debug=d password=1234".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_params">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-debug"
                            name="debug" value="d"
                        />
                        <nvpair id="S-instance_attributes-ip"
                            name="ip" value="i"
                        />
                        <nvpair id="S-instance_attributes-password"
                            name="password" value="1234"
                        />
                        <nvpair id="S-instance_attributes-username"
                            name="username" value="u"
                        />
                        <nvpair id="S-instance_attributes-verbose"
                            name="verbose" value="v"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            stderr_start=(
                "Warning: stonith option 'debug' is deprecated and might be "
                "removed in a future release, therefore it should not "
                "be used, use 'debug_file' instead\n"
            ),
        )

    def test_error_when_action_specified(self):
        self.assert_pcs_fail(
            "stonith create S fence_pcsmock_action action=reboot".split(),
            "Error: stonith option 'action' is deprecated and might be removed "
            "in a future release, therefore it should not be"
            " used, use 'pcmk_off_action', 'pcmk_reboot_action' instead, "
            "use --force to override\n" + ERRORS_HAVE_OCCURRED,
        )

    def test_warn_when_action_specified_forced(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_action action=reboot --force".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_action">
                    <instance_attributes id="S-instance_attributes">
                        <nvpair id="S-instance_attributes-action"
                            name="action" value="reboot"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            stderr_full=(
                "Warning: stonith option 'action' is deprecated and might be "
                "removed in a future release, therefore it should not be "
                "used, use 'pcmk_off_action', 'pcmk_reboot_action' instead\n"
            ),
        )


class WithMeta(ResourceTest):
    def setUp(self):
        super().setUp()
        self.pcs_runner.mock_settings = get_mock_settings()

    def test_simplest_with_meta_provides(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal meta provides=something".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-provides" name="provides"
                            value="something"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_base_with_agent_that_provides_unfencing_with_meta_provides(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_unfencing meta provides=something".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_unfencing">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-provides" name="provides"
                            value="unfencing"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_simplest_with_unknown_meta_attributes(self):
        self.assert_effect(
            (
                "stonith create S fence_pcsmock_minimal meta "
                "unknown_meta=unknown_value"
            ).split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-unknown_meta"
                            name="unknown_meta" value="unknown_value"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            stderr_full=fixture_meta_attributes_warning(
                ["unknown_meta"], ra_const.STONITH_META
            ),
        )


class InGroup(ResourceTest):
    def setUp(self):
        super().setUp()
        self.pcs_runner.mock_settings = get_mock_settings()

    def test_fail_create_in_group(self):
        self.assert_pcs_fail(
            "stonith create S fence_pcsmock_minimal --group G --after S1".split(),
            (
                "Error: Specified options '--after', '--group' are not "
                "supported in this command\n"
                "Hint: Syntax has changed from previous version. See 'man pcs' "
                "-> Changes in pcs-0.12.\n"
            ),
        )



import json
import shlex
from textwrap import dedent
from unittest import TestCase

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.resource.list import CibResourcesDto

from pcs_test.tier1.resource.test_config import ResourceConfigCmdMixin
from pcs_test.tools import resources_dto
from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

FIXTURE_JSON_WARNING = (
    "Warning: Fencing levels are not included because this command "
    "could only export stonith configuration previously. This cannot "
    "be changed to avoid breaking existing tooling. To export fencing "
    "levels, run 'pcs stonith level config --output-format=json'\n"
)


class StonithConfigJson(AssertPcsMixin, TestCase):
    def setUp(self):
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-all.xml"),
        )
        self.maxDiff = None

    def test_all(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["stonith", "config", "--output-format=json"],
        )
        self.assertEqual(stderr, FIXTURE_JSON_WARNING)
        self.assertEqual(retval, 0)
        expected = CibResourcesDto(
            primitives=[
                resources_dto.STONITH_S1,
                resources_dto.STONITH_S2,
            ],
            clones=[],
            groups=[],
            bundles=[],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))

    def test_get_specified(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["stonith", "config", "--output-format=json", "S1"],
        )
        self.assertEqual(stderr, FIXTURE_JSON_WARNING)
        self.assertEqual(retval, 0)
        expected = CibResourcesDto(
            primitives=[
                resources_dto.STONITH_S1,
            ],
            clones=[],
            groups=[],
            bundles=[],
        )
        self.assertEqual(json.loads(stdout), to_dict(expected))


class StonithConfigCmd(ResourceConfigCmdMixin, TestCase):
    sub_command = "stonith"
    expected_json_config_stderr = FIXTURE_JSON_WARNING

    @staticmethod
    def _get_tmp_file_name():
        return "tier1_stonith_test_config_cib.xml"

    def test_unsupported_stonith(self):
        self.pcs_runner_orig = PcsRunner(
            cib_file=get_test_resource("cib-unsupported-stonith-config.xml")
        )
        stdout, stderr, retval = self.pcs_runner_orig.run(
            ["stonith", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        self.assertEqual(stderr, "")
        self._run_commands(stdout)
        self.assertEqual(
            self._get_as_json(self.pcs_runner_new),
            self._get_as_json(self.pcs_runner_orig),
        )

    def test_plaintext(self):
        stdout, stderr, retval = self.pcs_runner_orig.run(
            [self.sub_command, "config"]
        )
        self.assertEqual(
            stdout,
            dedent(
                """\
                Resource: S1 (class=stonith type=fence_pcsmock_params)
                  Attributes: S1-instance_attributes
                    action=reboot
                    ip=203.0.113.1
                    username=testuser
                  Operations:
                    monitor: S1-monitor-interval-60s
                      interval=60s
                Resource: S2 (class=stonith type=fence_pcsmock_minimal)
                  Description: S2 description
                  Operations:
                    monitor: S2-monitor-interval-60s
                      interval=60s
                """
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)


class StonithConfigWithLevelsCmd(TestCase):
    def setUp(self):
        self.old_cib = get_tmp_file("tier1_stonith_cmd_old")
        write_file_to_tmpfile(
            get_test_resource("cib-fencing-levels.xml"), self.old_cib
        )
        self.new_cib = get_tmp_file("tier1_stonith_cmd_new")
        write_file_to_tmpfile(get_test_resource("cib-empty.xml"), self.new_cib)
        self.old_pcs_runner = PcsRunner(self.old_cib.name)
        self.new_pcs_runner = PcsRunner(self.new_cib.name)
        self.maxDiff = None

    def tearDown(self):
        self.old_cib.close()
        self.new_cib.close()

    def _get_stonith_resources_json(self, runner):
        stdout, stderr, retval = runner.run(
            ["stonith", "config", "--output-format=json"]
        )
        self.assertEqual(stderr, FIXTURE_JSON_WARNING)
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def _get_stonith_level_json(self, runner):
        stdout, stderr, retval = runner.run(
            ["stonith", "level", "config", "--output-format=json"]
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_success(self):
        stdout, stderr, retval = self.old_pcs_runner.run(
            ["stonith", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        commands = [
            shlex.split(command)[1:]
            for command in stdout.replace("\\\n", "").strip().split(";\n")
        ]

        for cmd in commands:
            stdout, stderr, retval = self.new_pcs_runner.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_stonith_resources_json(self.new_pcs_runner),
            self._get_stonith_resources_json(self.old_pcs_runner),
        )
        self.assertEqual(
            self._get_stonith_level_json(self.new_pcs_runner),
            self._get_stonith_level_json(self.old_pcs_runner),
        )


class StonithConfigText(AssertPcsMixin, TestCase):
    FIXTURE_S1 = (
        "Resource: S1 (class=stonith type=fence_pcsmock_minimal)\n"
        "  Operations:\n"
        "    monitor: S1-monitor-interval-60s\n"
        "      interval=60s\n"
    )
    FIXTURE_OTHER_RESOURCES = (
        "Resource: S2 (class=stonith type=fence_pcsmock_minimal)\n"
        "  Operations:\n"
        "    monitor: S2-monitor-interval-60s\n"
        "      interval=60s\n"
        "Resource: S3 (class=stonith type=fence_pcsmock_minimal)\n"
        "  Operations:\n"
        "    monitor: S3-monitor-interval-60s\n"
        "      interval=60s\n"
        "Resource: S4 (class=stonith type=fence_pcsmock_minimal)\n"
        "  Operations:\n"
        "    monitor: S4-monitor-interval-60s\n"
        "      interval=60s\n"
    )
    FIXTURE_FENCING_LEVELS = (
        "\n"
        "Fencing Levels:\n"
        "  Target (node): rh-1\n"
        "    Level 1: S1\n"
        "    Level 2: S2\n"
        "  Target (node): rh-2\n"
        "    Level 1: S3 S4\n"
        "  Target (regexp): rh-.*\n"
        "    Level 3: S4\n"
        "  Target (attribute): foo=bar\n"
        "    Level 4: S1 S2 S3\n"
    )
    FIXTURE_ALL = FIXTURE_S1 + FIXTURE_OTHER_RESOURCES + FIXTURE_FENCING_LEVELS

    def setUp(self):
        self.maxDiff = None
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-fencing-levels.xml")
        )

    def test_success(self):
        self.assert_pcs_success(
            ["stonith", "config", "--output-format=text"],
            stdout_full=self.FIXTURE_ALL,
        )

    def test_success_no_option(self):
        self.assert_pcs_success(
            ["stonith", "config"], stdout_full=self.FIXTURE_ALL
        )

    def test_success_filtered(self):
        self.assert_pcs_success(
            ["stonith", "config", "S1", "--output-format=text"],
            stdout_full=self.FIXTURE_S1 + self.FIXTURE_FENCING_LEVELS,
        )



import os.path
from dataclasses import dataclass
from typing import cast

from lxml import etree
from lxml.etree import _Element

from pcs import settings
from pcs.common.tools import xml_fromstring
from pcs.common.types import StringSequence


@dataclass(frozen=True)
class Status:
    code: int
    message: str
    errors: StringSequence


def get_api_result_dom(xml: str) -> _Element:
    # raises etree.XMLSyntaxError and etree.DocumentInvalid
    rng = settings.pacemaker_api_result_schema
    dom = xml_fromstring(xml)
    if os.path.isfile(rng):
        etree.RelaxNG(file=rng).assertValid(dom)
    return dom


def get_status_from_api_result(dom: _Element) -> Status:
    errors = []
    status_el = cast(_Element, dom.find("./status"))
    errors_el = status_el.find("errors")
    if errors_el is not None:
        errors = [
            str((error_el.text or "")).strip()
            for error_el in errors_el.iterfind("error")
        ]
    return Status(
        code=int(str(status_el.get("code"))),
        message=str(status_el.get("message")),
        errors=errors,
    )



import dataclasses

from pcs.lib.file.json import JsonExporter
from pcs.lib.interface.config import ExporterInterface

from .types import KnownHosts


class Exporter(ExporterInterface):
    @staticmethod
    def export(config_structure: KnownHosts) -> bytes:
        data = dataclasses.asdict(config_structure)

        # remove attribute not stored in file
        for host in data["known_hosts"].values():
            del host["name"]

        return JsonExporter.export(data)



import json
from typing import Any

from pcs.cli.cluster_property.output import (
    PropertyConfigurationFacade,
    cluster_property_metadata_to_text,
    properties_defaults_to_text,
    properties_to_cmd,
    properties_to_text,
    properties_to_text_with_default_mark,
)
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import smart_wrap_text
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_OPTION,
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    OUTPUT_FORMAT_VALUE_TEXT,
    Argv,
    InputModifiers,
    KeyValueParser,
    ensure_unique_args,
)
from pcs.cli.reports.output import deprecation_warning
from pcs.common import reports
from pcs.common.interface import dto
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.pacemaker.nvset import ListCibNvsetDto
from pcs.common.str_tools import format_list, format_plural


def set_property(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - allow unknown options
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    if not argv:
        raise CmdLineInputError()
    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)
    cluster_options = KeyValueParser(argv).get_unique()
    lib.cluster_property.set_properties(cluster_options, force_flags)


def unset_property(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - no error when removing not existing properties
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    if not argv:
        raise CmdLineInputError()
    force_flags = set()
    if modifiers.get("--force"):
        force_flags.add(reports.codes.FORCE)
    else:
        ensure_unique_args(argv)

    lib.cluster_property.set_properties(dict.fromkeys(argv, ""), force_flags)


def config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --all - list configured properties with values and properties with
          default values if not in configuration
      * --defaults - list only default values of properties, only properties
          with a default value are listed
      * --output-format - supported formats: text, cmd, json
    """
    modifiers.ensure_only_supported(
        "-f", "--all", "--defaults", output_format_supported=True
    )
    mutually_exclusive_options = ["--all", "--defaults", "--output-format"]
    if argv and modifiers.is_specified_any(mutually_exclusive_options):
        raise CmdLineInputError(
            "cannot specify properties when using {}".format(
                format_list(mutually_exclusive_options)
            )
        )
    modifiers.ensure_not_mutually_exclusive(*mutually_exclusive_options)
    output_format = modifiers.get_output_format()

    if (
        argv
        or output_format == OUTPUT_FORMAT_VALUE_CMD
        or modifiers.get("--all")
    ):
        properties_facade = PropertyConfigurationFacade.from_properties_dtos(
            lib.cluster_property.get_properties(),
            lib.cluster_property.get_properties_metadata(),
        )
    elif modifiers.get("--defaults"):
        # do not load set properties
        # --defaults should work without a cib file
        properties_facade = (
            PropertyConfigurationFacade.from_properties_metadata(
                lib.cluster_property.get_properties_metadata()
            )
        )
    else:
        # json or default text
        # do not load properties metadata, only configured properties are needed
        properties_facade = PropertyConfigurationFacade.from_properties_config(
            lib.cluster_property.get_properties()
        )

    if argv:
        output = "\n".join(
            properties_to_text_with_default_mark(
                properties_facade, property_names=argv
            )
        )
    elif modifiers.get("--all"):
        output = "\n".join(
            properties_to_text_with_default_mark(properties_facade)
        )
    elif modifiers.get("--defaults"):
        deprecation_warning(
            "Option --defaults is deprecated and will be removed. "
            "Please use command 'pcs property defaults' instead."
        )
        output = "\n".join(
            properties_defaults_to_text(
                properties_facade.get_defaults(include_advanced=True)
            )
        )
    elif output_format == OUTPUT_FORMAT_VALUE_CMD:
        output = " \\\n".join(properties_to_cmd(properties_facade))
    elif output_format == OUTPUT_FORMAT_VALUE_JSON:
        output = json.dumps(
            dto.to_dict(ListCibNvsetDto(properties_facade.properties[0:1]))
        )
    else:
        output = "\n".join(properties_to_text(properties_facade))

    if output:
        print(output)


def defaults(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --full - also list advanced cluster properties
    """
    modifiers.ensure_only_supported("--full")
    if argv and modifiers.is_specified("--full"):
        raise CmdLineInputError("cannot specify properties when using '--full'")
    properties_facade = PropertyConfigurationFacade.from_properties_metadata(
        lib.cluster_property.get_properties_metadata()
    )
    defaults_dict = properties_facade.get_defaults(
        argv, include_advanced=modifiers.is_specified("--full")
    )
    extra_args = set(argv) - defaults_dict.keys()
    if extra_args:
        raise CmdLineInputError(
            "No default value for {property_pl}: {name_list}".format(
                property_pl=format_plural(extra_args, "property"),
                name_list=format_list(list(extra_args)),
            )
        )
    output = "\n".join(properties_defaults_to_text(defaults_dict))
    if output:
        print(output)


def describe(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --full - also list advanced cluster properties
      * --output-format - supported formats: text, json
    """
    modifiers.ensure_only_supported("--full", output_format_supported=True)
    if argv and modifiers.is_specified("--full"):
        raise CmdLineInputError("cannot specify properties when using '--full'")
    output_format = modifiers.get_output_format(
        supported_formats={OUTPUT_FORMAT_VALUE_TEXT, OUTPUT_FORMAT_VALUE_JSON}
    )
    if output_format == OUTPUT_FORMAT_VALUE_JSON and (
        argv or modifiers.is_specified("--full")
    ):
        raise CmdLineInputError(
            "property filtering is not supported with "
            f"{OUTPUT_FORMAT_OPTION}={OUTPUT_FORMAT_VALUE_JSON}"
        )
    properties_facade = PropertyConfigurationFacade.from_properties_metadata(
        lib.cluster_property.get_properties_metadata()
    )
    if output_format == OUTPUT_FORMAT_VALUE_JSON:
        output = json.dumps(
            dto.to_dict(
                ClusterPropertyMetadataDto(
                    properties_metadata=properties_facade.properties_metadata,
                    readonly_properties=properties_facade.readonly_properties,
                )
            )
        )
    else:
        filtered_metadata = properties_facade.get_properties_metadata(
            argv, include_advanced=modifiers.is_specified("--full")
        )
        extra_args = set(argv) - {
            metadata.name for metadata in filtered_metadata
        }
        if extra_args:
            raise CmdLineInputError(
                "No description for {property_pl}: {name_list}".format(
                    property_pl=format_plural(extra_args, "property"),
                    name_list=format_list(list(extra_args)),
                )
            )
        output = "\n".join(
            smart_wrap_text(
                cluster_property_metadata_to_text(
                    sorted(filtered_metadata, key=lambda x: x.name)
                )
            )
        )
    if output:
        print(output)


def print_cluster_properties_definition_legacy(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    print(
        json.dumps(
            lib.cluster_property.get_cluster_properties_definition_legacy()
        )
    )



from textwrap import dedent
from unittest import TestCase

from pcs.cli.cluster_property import output as cluster_property
from pcs.cli.cluster_property.output import PropertyConfigurationFacade
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
    ListCibNvsetDto,
)
from pcs.common.resource_agent.dto import ResourceAgentParameterDto

FIXTURE_TWO_PROPERTY_SETS = [
    CibNvsetDto(
        id="id1",
        options={"score": "150"},
        rule=None,
        nvpairs=[
            CibNvpairDto(id="", name="readonly1", value="ro_val1"),
            CibNvpairDto(id="", name="readonly2", value="ro_val2"),
            CibNvpairDto(id="", name="property2", value="val2"),
            CibNvpairDto(id="", name="property1", value="val1"),
            CibNvpairDto(id="", name="property1", value="duplicate_val1"),
        ],
    ),
    CibNvsetDto(
        id="id2",
        options={"score": "100"},
        rule=None,
        nvpairs=[
            CibNvpairDto(id="", name="readonly3", value="ro_val3"),
            CibNvpairDto(id="", name="property3", value="val3"),
        ],
    ),
]

FIXTURE_READONLY_PROPERTIES_LIST = ["readonly1", "readonly2"]

FIXTURE_TEXT_OUTPUT_FIRST_SET = dedent(
    """\
    Cluster Properties: id1 score=150
      property1=duplicate_val1
      property1=val1
      property2=val2
      readonly1=ro_val1
      readonly2=ro_val2
    """
)


def fixture_property_metadata(
    name="property-name",
    shortdesc=None,
    longdesc=None,
    type="string",  # noqa: A002
    default=None,
    enum_values=None,
    advanced=False,
):
    # pylint: disable=redefined-builtin
    return ResourceAgentParameterDto(
        name=name,
        shortdesc=shortdesc,
        longdesc=longdesc,
        type=type,
        default=default,
        enum_values=enum_values,
        required=False,
        advanced=advanced,
        deprecated=False,
        deprecated_by=[],
        deprecated_desc=None,
        unique_group=None,
        reloadable=False,
    )


FIXTURE_PROPERTY_METADATA_LIST = [
    fixture_property_metadata(name="property1", default="duplicate_default1"),
    fixture_property_metadata(name="property1", default="default1"),
    fixture_property_metadata(name="property2", default="default2"),
    fixture_property_metadata(
        name="property3", default="default3", advanced=True
    ),
    fixture_property_metadata(
        name="property4", default="default4", advanced=True
    ),
]


class TestPropertyConfigurationFacadeCreate(TestCase):
    def test_from_properties_dtos(self):
        facade = PropertyConfigurationFacade.from_properties_dtos(
            properties_dto=ListCibNvsetDto(nvsets=FIXTURE_TWO_PROPERTY_SETS),
            properties_metadata_dto=ClusterPropertyMetadataDto(
                properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
                readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
            ),
        )
        self.assertEqual(facade.properties, FIXTURE_TWO_PROPERTY_SETS)
        self.assertEqual(
            facade.properties_metadata, FIXTURE_PROPERTY_METADATA_LIST
        )
        self.assertEqual(
            facade.readonly_properties, FIXTURE_READONLY_PROPERTIES_LIST
        )

    def test_from_properties_config(self):
        facade = PropertyConfigurationFacade.from_properties_config(
            properties_dto=ListCibNvsetDto(nvsets=FIXTURE_TWO_PROPERTY_SETS)
        )
        self.assertEqual(facade.properties, FIXTURE_TWO_PROPERTY_SETS)
        self.assertEqual(facade.properties_metadata, [])
        self.assertEqual(facade.readonly_properties, [])

    def test_from_properties_metadata(self):
        facade = PropertyConfigurationFacade.from_properties_metadata(
            properties_metadata_dto=ClusterPropertyMetadataDto(
                properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
                readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
            ),
        )
        self.assertEqual(facade.properties, [])
        self.assertEqual(
            facade.properties_metadata, FIXTURE_PROPERTY_METADATA_LIST
        )
        self.assertEqual(
            facade.readonly_properties, FIXTURE_READONLY_PROPERTIES_LIST
        )


class TestPropertyConfigurationFacadeGetPropertyValue(TestCase):
    def setUp(self):
        self.facade = PropertyConfigurationFacade(
            properties=FIXTURE_TWO_PROPERTY_SETS,
            properties_metadata=[],
            readonly_properties=[],
        )

    def test_property_value_from_first_set(self):
        self.assertEqual(self.facade.get_property_value("property2"), "val2")

    def test_property_value_from_second_set(self):
        self.assertEqual(self.facade.get_property_value("property3"), None)

    def test_property_value_not_in_set(self):
        self.assertEqual(self.facade.get_property_value("not-there"), None)

    def test_custom_default(self):
        self.assertEqual(
            self.facade.get_property_value(
                "not-there", custom_default="custom"
            ),
            "custom",
        )

    def test_property_with_multiple_values(self):
        self.assertEqual(
            self.facade.get_property_value("property1"), "duplicate_val1"
        )


class TestPropertyConfigurationFacadeGetPropertyValueOrDefault(TestCase):
    def setUp(self):
        self.facade = PropertyConfigurationFacade(
            properties=FIXTURE_TWO_PROPERTY_SETS,
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
        )

    def test_property_value_from_first_set(self):
        self.assertEqual(
            self.facade.get_property_value_or_default("property2"), "val2"
        )

    def test_property_value_not_in_set(self):
        self.assertEqual(
            self.facade.get_property_value_or_default("property3"), "default3"
        )

    def test_custom_default_with_existing_default(self):
        self.assertEqual(
            self.facade.get_property_value_or_default(
                "property3", custom_default="custom"
            ),
            "default3",
        )

    def test_custom_default_with_not_existing_default(self):
        self.assertEqual(
            self.facade.get_property_value_or_default(
                "not-there", custom_default="custom"
            ),
            "custom",
        )


class TestPropertyConfigurationFacadeGetDefaults(TestCase):
    def setUp(self):
        self.facade = PropertyConfigurationFacade(
            properties=[],
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
        )

    def test_defaults_not_advanced(self):
        result_dict = {"property1": "default1", "property2": "default2"}
        self.assertEqual(self.facade.get_defaults(), result_dict)

    def test_defaults_advanced(self):
        result_dict = {
            "property1": "default1",
            "property2": "default2",
            "property3": "default3",
            "property4": "default4",
        }
        self.assertEqual(
            self.facade.get_defaults(include_advanced=True), result_dict
        )

    def test_specified(self):
        result_dict = {"property1": "default1", "property4": "default4"}
        self.assertEqual(
            self.facade.get_defaults(
                property_names=["property4", "property1", "nodefault1"]
            ),
            result_dict,
        )

    def test_properties_without_defaults(self):
        result_dict = {}
        self.assertEqual(
            self.facade.get_defaults(
                property_names=["nodefault1", "nodefault2"]
            ),
            result_dict,
        )


class TestPropertyConfigurationFacadeGetPropertiesMetadata(TestCase):
    def setUp(self):
        self.facade = PropertyConfigurationFacade(
            properties=[],
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=[],
        )

    def test_metadata_without_advanced(self):
        metadata = FIXTURE_PROPERTY_METADATA_LIST[1:3]
        self.assertCountEqual(self.facade.get_properties_metadata(), metadata)

    def test_metadata_with_advanced(self):
        metadata = FIXTURE_PROPERTY_METADATA_LIST[1:]
        self.assertCountEqual(
            self.facade.get_properties_metadata(include_advanced=True),
            metadata,
        )

    def test_metadata_specified(self):
        metadata = (
            FIXTURE_PROPERTY_METADATA_LIST[1:2]
            + FIXTURE_PROPERTY_METADATA_LIST[-1:]
        )
        self.assertCountEqual(
            self.facade.get_properties_metadata(
                property_names=["property4", "property1"]
            ),
            metadata,
        )


class TestPropertyConfigurationFacadeGetNameValueDefaultList(TestCase):
    def setUp(self):
        self.facade = PropertyConfigurationFacade(
            properties=FIXTURE_TWO_PROPERTY_SETS,
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
        )

    def test_get_name_value_default_list(self):
        tuple_list = [
            ("readonly1", "ro_val1", False),
            ("readonly2", "ro_val2", False),
            ("property2", "val2", False),
            ("property1", "val1", False),
            ("property1", "duplicate_val1", False),
            ("property3", "default3", True),
            ("property4", "default4", True),
        ]
        self.assertEqual(self.facade.get_name_value_default_list(), tuple_list)


class TestPropertiesToText(TestCase):
    def assert_lines(self, facade, output):
        self.assertEqual(
            "\n".join(cluster_property.properties_to_text(facade)) + "\n",
            output,
        )

    def test_no_cluster_properties(self):
        facade = PropertyConfigurationFacade(
            properties=[], properties_metadata=[], readonly_properties=[]
        )
        output = "\n"
        self.assert_lines(facade, output)

    def test_empty_cluster_property_set(self):
        facade = PropertyConfigurationFacade(
            properties=[
                CibNvsetDto(id="id1", options={}, rule=None, nvpairs=[])
            ],
            properties_metadata=[],
            readonly_properties=[],
        )
        output = dedent(
            """\
            Cluster Properties: id1
            """
        )
        self.assert_lines(facade, output)

    def test_one_cluster_property_set(self):
        facade = PropertyConfigurationFacade(
            properties=FIXTURE_TWO_PROPERTY_SETS[0:1],
            properties_metadata=[],
            readonly_properties=[],
        )
        output = FIXTURE_TEXT_OUTPUT_FIRST_SET
        self.assert_lines(facade, output)

    def test_more_cluster_property_sets_first_is_displayed(self):
        facade = PropertyConfigurationFacade(
            properties=FIXTURE_TWO_PROPERTY_SETS,
            properties_metadata=[],
            readonly_properties=[],
        )
        output = FIXTURE_TEXT_OUTPUT_FIRST_SET
        self.assert_lines(facade, output)


class TestPropertiesToTextWithDefaultMark(TestCase):
    def assert_lines(self, facade, output, property_names=None):
        self.assertEqual(
            "\n".join(
                cluster_property.properties_to_text_with_default_mark(
                    facade, property_names=property_names
                )
            )
            + "\n",
            output,
        )

    def test_no_cluster_properties_and_no_defaults(self):
        facade = PropertyConfigurationFacade(
            properties=[], properties_metadata=[], readonly_properties=[]
        )
        output = dedent(
            """\
            Cluster Properties:
            """
        )
        self.assert_lines(facade, output)

    def test_no_cluster_properties_and_defaults_only(self):
        facade = PropertyConfigurationFacade(
            properties=[],
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=[],
        )
        output = dedent(
            """\
            Cluster Properties:
              property1=default1 (default)
              property2=default2 (default)
              property3=default3 (default)
              property4=default4 (default)
            """
        )
        self.assert_lines(facade, output)

    def test_configured_properties_and_properties_with_defaults(self):
        facade = PropertyConfigurationFacade(
            properties=[
                CibNvsetDto(
                    id="id1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="", name="readonly1", value="ro_val1"),
                        CibNvpairDto(id="", name="readonly2", value="ro_val2"),
                        CibNvpairDto(id="", name="property2", value="val2"),
                        CibNvpairDto(id="", name="property1", value="default1"),
                    ],
                )
            ],
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
        )
        output = dedent(
            """\
            Cluster Properties: id1
              property1=default1
              property2=val2
              property3=default3 (default)
              property4=default4 (default)
              readonly1=ro_val1
              readonly2=ro_val2
            """
        )
        self.assert_lines(facade, output)

    def test_specified_properties(self):
        facade = PropertyConfigurationFacade(
            properties=[
                CibNvsetDto(
                    id="id1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="", name="readonly1", value="ro_val1"),
                        CibNvpairDto(id="", name="readonly2", value="ro_val2"),
                        CibNvpairDto(id="", name="property2", value="val2"),
                        CibNvpairDto(id="", name="property1", value="default1"),
                    ],
                )
            ],
            properties_metadata=FIXTURE_PROPERTY_METADATA_LIST,
            readonly_properties=FIXTURE_READONLY_PROPERTIES_LIST,
        )
        output = dedent(
            """\
            Cluster Properties: id1
              property2=val2
              property4=default4 (default)
              readonly2=ro_val2
            """
        )
        self.assert_lines(
            facade,
            output,
            property_names=["property2", "readonly2", "property4", "other"],
        )


class TestPropertiesToCmd(TestCase):
    def assert_lines(self, facade, output):
        self.assertEqual(
            " \\\n".join(cluster_property.properties_to_cmd(facade)) + "\n",
            output,
        )

    def test_no_cluster_properties(self):
        facade = PropertyConfigurationFacade(
            properties=[], properties_metadata=[], readonly_properties=[]
        )
        output = "\n"
        self.assert_lines(facade, output)

    def test_only_readonly_properties(self):
        facade = PropertyConfigurationFacade(
            properties=[
                CibNvsetDto(
                    id="id1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="", name="readonly1", value="ro_val1"),
                        CibNvpairDto(id="", name="readonly2", value="ro_val2"),
                    ],
                )
            ],
            properties_metadata=[],
            readonly_properties=["readonly1", "readonly2"],
        )
        output = "\n"
        self.assert_lines(facade, output)

    def test_properties_cmd_without_readonly_properties(self):
        facade = PropertyConfigurationFacade(
            properties=[
                CibNvsetDto(
                    id="id1",
                    options={},
                    rule=None,
                    nvpairs=[
                        CibNvpairDto(id="", name="readonly1", value="ro_val1"),
                        CibNvpairDto(id="", name="readonly2", value="ro_val2"),
                        CibNvpairDto(id="", name="property2", value="val2"),
                        CibNvpairDto(id="", name="property1", value="val1"),
                    ],
                )
            ],
            properties_metadata=[],
            readonly_properties=["readonly1", "readonly2"],
        )
        output = dedent(
            """\
            pcs property set --force -- \\
              property2=val2 \\
              property1=val1
            """
        )
        self.assert_lines(facade, output)

    def test_only_first_set_is_supported_for_cmd_output(self):
        facade = PropertyConfigurationFacade(
            properties=FIXTURE_TWO_PROPERTY_SETS,
            properties_metadata=[],
            readonly_properties=["readonly1", "readonly2"],
        )
        output = dedent(
            """\
            pcs property set --force -- \\
              property2=val2 \\
              property1=val1 \\
              property1=duplicate_val1
            """
        )
        self.assert_lines(facade, output)


class TestPropertiesDefaultsToText(TestCase):
    def assert_lines(self, defaults_dict, lines):
        self.assertEqual(
            "\n".join(
                cluster_property.properties_defaults_to_text(defaults_dict)
            )
            + "\n",
            lines,
        )

    def test_no_defaults(self):
        defaults = {}
        output = "\n"
        self.assert_lines(defaults, output)

    def test_some_defaults(self):
        defaults = {"def1": "val1", "def2": "val2", "def3": "val3"}
        output = dedent(
            """\
            def1=val1
            def2=val2
            def3=val3
            """
        )
        self.assert_lines(defaults, output)


class TestClusterPropertyMetadataToText(TestCase):
    def assert_lines(self, metadata, lines):
        self.assertEqual(
            "\n".join(
                cluster_property.cluster_property_metadata_to_text(metadata)
            )
            + "\n",
            lines,
        )

    def test_multiple_properties_metadata(self):
        metadata = [
            fixture_property_metadata(
                name="property-a",
                shortdesc="desc",
                type="percentage",
                default="80%",
                advanced=True,
            ),
            fixture_property_metadata(
                name="property-b",
                type="some type",
                enum_values=["a", "b"],
                default="a",
            ),
        ]
        output = dedent(
            """\
            property-a (advanced use only)
              Description: desc
              Type: percentage
              Default: 80%
            property-b
              Description: No description available
              Allowed values: 'a', 'b'
              Default: a
            """
        )
        self.assert_lines(metadata, output)

    def test_empty_metadata(self):
        metadata = []
        output = "\n"
        self.assert_lines(metadata, output)

    def test_no_longdesc_and_shortdesc_defined(self):
        metadata = [fixture_property_metadata()]
        output = dedent(
            """\
            property-name
              Description: No description available
              Type: string
            """
        )
        self.assert_lines(metadata, output)

    def test_longdesc_and_shortdesc_defined(self):
        metadata = [
            fixture_property_metadata(
                shortdesc="'\nshort\ndesc\n'", longdesc="'\nlong\ndesc\n'"
            )
        ]
        output = dedent(
            """\
            property-name
              Description: ' long desc '
              Type: string
            """
        )
        self.assert_lines(metadata, output)

    def test_shortdesc_only(self):
        metadata = [fixture_property_metadata(shortdesc="'\nshort\ndesc\n'")]
        output = dedent(
            """\
            property-name
              Description: ' short desc '
              Type: string
            """
        )
        self.assert_lines(metadata, output)

    def test_enum_values(self):
        metadata = [
            fixture_property_metadata(
                type="select", enum_values=["a", "b", "c"]
            )
        ]
        output = dedent(
            """\
            property-name
              Description: No description available
              Allowed values: 'a', 'b', 'c'
            """
        )
        self.assert_lines(metadata, output)

    def test_default_value(self):
        metadata = [fixture_property_metadata(default="default_value")]
        output = dedent(
            """\
            property-name
              Description: No description available
              Type: string
              Default: default_value
            """
        )
        self.assert_lines(metadata, output)

    def test_advanced_value(self):
        metadata = [fixture_property_metadata(advanced=True)]
        output = dedent(
            """\
            property-name (advanced use only)
              Description: No description available
              Type: string
            """
        )
        self.assert_lines(metadata, output)



from typing import (
    Any,
    Optional,
    cast,
)

from pcs.cli.common.output import INDENT_STEP
from pcs.cli.common.tools import print_to_stderr
from pcs.cli.constraint import output
from pcs.common import reports
from pcs.common.pacemaker.constraint import CibConstraintsDto
from pcs.common.str_tools import (
    format_list,
    indent,
)
from pcs.common.types import StringIterable
from pcs.lib.errors import LibraryError

from .output import process_library_reports
from .processor import ReportItemPreprocessor


def get_duplicate_constraint_exists_preprocessor(
    lib: Any,
) -> ReportItemPreprocessor:
    constraints_dto: Optional[CibConstraintsDto] = None

    def _report_item_preprocessor(  # noqa: PLR0912
        report_item: reports.ReportItem,
    ) -> Optional[reports.ReportItem]:
        """
        Provide additional info based on DuplicateConstraintsExist message

        Drop deprecated DuplicateConstraintsList message. This message
        contained structured info about duplicate constraints.
        Intercept DuplicateConstraintsExist message and extract constraint IDs
        from it. Load constraints from CIB using library and print those
        matching IDs from the message.
        """

        # pylint: disable=too-many-branches
        nonlocal constraints_dto

        def my_print(lines: StringIterable) -> None:
            print_to_stderr("\n".join(indent(lines, INDENT_STEP)))

        if (
            report_item.message.code
            == reports.deprecated_codes.DUPLICATE_CONSTRAINTS_LIST
        ):
            return None

        if isinstance(
            report_item.message, reports.messages.DuplicateConstraintsExist
        ):
            duplicate_id_list = report_item.message.constraint_ids
            try:
                if constraints_dto is None:
                    constraints_dto = cast(
                        CibConstraintsDto,
                        lib.constraint.get_config(evaluate_rules=False),
                    )
                print_to_stderr("Duplicate constraints:")
                for dto_lp in constraints_dto.location:
                    if dto_lp.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.location.plain_constraint_to_text(
                                dto_lp, True
                            )
                        )
                for dto_ls in constraints_dto.location_set:
                    if dto_ls.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.location.set_constraint_to_text(dto_ls, True)
                        )
                for dto_cp in constraints_dto.colocation:
                    if dto_cp.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.colocation.plain_constraint_to_text(
                                dto_cp, True
                            )
                        )
                for dto_cs in constraints_dto.colocation_set:
                    if dto_cs.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.colocation.set_constraint_to_text(
                                dto_cs, True
                            )
                        )
                for dto_op in constraints_dto.order:
                    if dto_op.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.order.plain_constraint_to_text(dto_op, True)
                        )
                for dto_os in constraints_dto.order_set:
                    if dto_os.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.order.set_constraint_to_text(dto_os, True)
                        )
                for dto_tp in constraints_dto.ticket:
                    if dto_tp.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.ticket.plain_constraint_to_text(dto_tp, True)
                        )
                for dto_ts in constraints_dto.ticket_set:
                    if dto_ts.attributes.constraint_id in duplicate_id_list:
                        my_print(
                            output.ticket.set_constraint_to_text(dto_ts, True)
                        )
            except LibraryError as e:
                # If reading constraints failed, print their IDs and ignore the
                # exception. We want to print remaining reports from the
                # originally called command. Raising the exception would
                # prevent that. Also, it is not correct to exit with an error
                # just because we were unable to get optional additional info
                # for printing a report.
                if e.output:
                    print_to_stderr(e.output)
                if e.args:
                    process_library_reports(
                        cast(reports.ReportItemList, e.args),
                        exit_on_error=False,
                    )
                print_to_stderr(
                    "Duplicate constraints: " + format_list(duplicate_id_list)
                )

        return report_item

    return _report_item_preprocessor



def is_negative_score(score: str) -> bool:
    return score.startswith("-")


def abs_score(score: str) -> str:
    """
    return absolute value of score
    """
    if is_negative_score(score):
        return score[1:]
    return score



from typing import (
    Any,
    Optional,
)

from pcs.common.fencing_topology import (
    FencingTargetType,
    FencingTargetValue,
)
from pcs.common.pacemaker.fencing_topology import CibFencingTopologyDto
from pcs.common.types import StringSequence
from pcs.lib.cib import fencing_topology as cib_fencing_topology
from pcs.lib.cib.tools import get_fencing_topology
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.state import ClusterState


def add_level(
    lib_env: LibraryEnvironment,
    level: str,
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
    devices: StringSequence,
    level_id: Optional[str] = None,
    force_device: bool = False,
    force_node: bool = False,
) -> None:
    """
    Validate and add a new fencing level

    lib_env -- environment
    level -- level (index) of the new fencing level
    target_type -- the new fencing level target value type
    target_value -- the new fencing level target value
    devices -- list of stonith devices for the new fencing level
    level_id -- user specified id for the level element
    force_device -- continue even if a stonith device does not exist
    force_node -- continue even if a node (target) does not exist
    """
    cib = lib_env.get_cib()
    cib_fencing_topology.add_level(
        lib_env.report_processor,
        cib,
        level,
        target_type,
        target_value,
        devices,
        ClusterState(lib_env.get_cluster_state()).node_section.nodes,
        level_id,
        force_device,
        force_node,
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()
    lib_env.push_cib()


# DEPRECATED, use get_config_dto
def get_config(lib_env: LibraryEnvironment) -> list[dict[str, Any]]:
    """
    Get fencing levels configuration.

    Return a list of levels where each level is a dict with keys: target_type,
    target_value, level and devices. Devices is a list of stonith device ids.

    lib_env -- environment
    """
    cib = lib_env.get_cib()
    return cib_fencing_topology.export(get_fencing_topology(cib))


def get_config_dto(env: LibraryEnvironment) -> CibFencingTopologyDto:
    """
    Get fencing level configuration.

    env -- library environment
    """
    return cib_fencing_topology.fencing_topology_el_to_dto(
        get_fencing_topology(env.get_cib())
    )


def remove_all_levels(lib_env: LibraryEnvironment) -> None:
    """
    Remove all fencing levels

    lib_env -- environment
    """
    cib_fencing_topology.remove_all_levels(
        get_fencing_topology(lib_env.get_cib())
    )
    lib_env.push_cib()


def remove_levels_by_params(
    lib_env: LibraryEnvironment,
    level: Optional[str] = None,
    target_type: Optional[FencingTargetType] = None,
    target_value: Optional[FencingTargetValue] = None,
    devices: Optional[StringSequence] = None,
) -> None:
    """
    Remove specified fencing level(s).

    lib_env -- environment
    level -- level (index) of the fencing level to remove
    target_type -- the removed fencing level target value type
    target_value -- the removed fencing level target value
    devices -- list of stonith devices of the removed fencing level
    """
    topology_el = get_fencing_topology(lib_env.get_cib())
    report_list = cib_fencing_topology.remove_levels_by_params(
        topology_el,
        level,
        target_type,
        target_value,
        devices,
    )
    if lib_env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    lib_env.push_cib()


def verify(lib_env: LibraryEnvironment) -> None:
    """
    Check if all cluster nodes and stonith devices used in fencing levels exist

    lib_env -- environment
    """
    cib = lib_env.get_cib()
    lib_env.report_processor.report_list(
        cib_fencing_topology.verify(
            cib,
            ClusterState(lib_env.get_cluster_state()).node_section.nodes,
        )
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()
